# This contains logic for making "mapped" ops for all regularly declared ops.
# These derived ops are not yet fully Weave serializable due to some non-json stuff
#     in their closure. So we disable legacy_strict_op_saving when deriving them.

import copy
import inspect
import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import (
    storage,
    weave_internal,
    errors,
    parallelism,
    registry_mem,
    box,
    context_state,
    execute_fast,
    graph,
    op_args,
    op_def,
    op_policy,
)
from weave.legacy.weave.language_features.tagging import tag_store

USE_PARALLEL_DOWNLOAD = True
USE_PARALLEL_REFINE = True
USE_PARALLEL_RESOLVE = True

_PARALLEL_ALLOWLIST = ["op-refine_history_metrics", "artifactVersion-historyMetrics"]
PARALLEL_REFINE_ALLOWLIST = _PARALLEL_ALLOWLIST
PARALLEL_RESOLVE_ALLOWLIST = _PARALLEL_ALLOWLIST


class DeriveOpHandler:
    """
    Subclassing this class will add a new type of derived op to the Weave1 system
    """

    handler_id: typing.ClassVar[str]

    @staticmethod
    def derived_name(name: str) -> str:
        raise NotImplementedError()

    @staticmethod
    def should_derive_op(orig_op: op_def.OpDef) -> bool:
        raise NotImplementedError()

    @staticmethod
    def make_derived_op(orig_op: op_def.OpDef) -> op_def.OpDef:
        raise NotImplementedError()

    @staticmethod
    def handle_class_decorator_update(
        derived_op: op_def.OpDef,
        base_weave_type: type[types.Type],
        orig_op_new_name: str,
    ):
        raise NotImplementedError()


# These are a list of type names that should not be mapped
disallow_mapping_type_name_list = [
    "type",
    "list",
    "WeaveNDArray",
    "ArrowArray",
    "ArrowTable",
    "ArrowWeaveList",
    "dataframe",
    "sqltable",
    "projectArtifactVersions",
    "runs",
    "artifacts",
    "projectArtifactTypes",
    "invalid",
    "unknown",
    "none",
    "any",
    "groupresult",
    "dataframeTable",
    "ArrowWeaveList",
]


# This class implements a nullable mappable derived op handler. It will create a new op
# that is capable of handling a List<Optional<T>> input for a given op that takes a T
class MappedDeriveOpHandler(DeriveOpHandler):
    handler_id = "mapped"

    @staticmethod
    def derived_name(name: str) -> str:
        if "-" not in name:
            return f"mapped-{name}"
        else:
            return f"mapped_{name}"

    @staticmethod
    def should_derive_op(orig_op: op_def.OpDef) -> bool:
        """Returns True if the op should be mapped to a list of inputs."""
        named_args = orig_op.input_type.named_args()

        # The argument list must be named AND have at least 1 argument.
        if len(named_args) == 0:
            return False

        first_arg = named_args[0]

        # Enforce the disallow list
        if first_arg.type.name in disallow_mapping_type_name_list:
            return False

        # Here, we check if the first_arg is unknown. If it is, then we cannot tell
        # if it is supposed to be mapped.
        if first_arg.type == types.UnknownType():
            return False

        # If the first argument can be assigned to a list<any>, then we should not map it -
        # it will create unresolvable ambiguity between the current op and the mapped.
        if types.List(types.Any()).assign_type(first_arg.type):
            return False

        # If the first argument can be assigned a list of the first argument, then we should not map -
        # this too will create unresolvable ambiguity.
        if first_arg.type.assign_type(types.List(first_arg.type)):
            return False

        return True

    @staticmethod
    def make_derived_op(orig_op: op_def.OpDef) -> op_def.OpDef:
        mapped_op_name = MappedDeriveOpHandler.derived_name(
            orig_op.name
        )  # TODO: doesn't handle fqn
        named_args = orig_op.input_type.named_args()

        if len(named_args) == 0 or not isinstance(
            orig_op.input_type, op_args.OpNamedArgs
        ):
            raise errors.WeaveInternalError(
                f"Cannot make mapped op for op {orig_op.name} with no first named argument."
            )
        first_arg = named_args[0]
        mapped_param_name = first_arg.name

        output_type: typing.Union[types.Type, typing.Callable]
        # TODO: In all of these, we use types.optional, but this should only be
        # used if the input_type is also optional. However we don't have a
        # weave-way to check that yet :(.
        if not callable(orig_op.output_type):

            def new_output_type(input_type):
                object_type = input_type[mapped_param_name].object_type
                if types.is_optional(object_type):
                    return types.List(types.optional(object_type))
                return types.List(object_type)

            output_type = new_output_type
        else:

            def make_output_type(input_types):
                replacement = input_types[mapped_param_name].object_type

                # Remove the nulls from the inner type
                if types.is_optional(replacement):
                    replacement = types.non_none(replacement)
                if isinstance(replacement, types.NoneType):
                    return types.List(types.NoneType())

                # This is a special circumstance (aka "God Mode") where we are
                # inferring when an external caller is trying to weaveify this
                # function. In this specific case, we need to manually construct the
                # output_type. The main reason for this is that `merge` is not yet
                # implemented as a core op that looks the same in python and weave.
                # Therefore the `inner_input_types[mapped_param_name] = replacement`
                # line below will not work. This is a temporary fix until we can
                # implement `merge` as a core op.
                currently_weavifying = isinstance(
                    input_types, graph.Node
                ) and types.TypedDict({}).assign_type(input_types.type)
                if currently_weavifying:
                    op_dict = registry_mem.memory_registry.get_op("dict")
                    op_dict.instance = None
                    inner_input_types = input_types.merge(
                        op_dict(**{mapped_param_name: replacement})
                    )
                    try:
                        inner_res = orig_op.output_type(inner_input_types)
                    except errors.WeaveExpectedConstError as e:
                        raise errors.WeaveMakeFunctionError(
                            "function body expected const node."
                        )
                    if not isinstance(inner_res, graph.Node):
                        raise errors.WeaveMakeFunctionError(
                            "output_type function must return a node."
                        )
                    from weave.legacy.weave.ops_primitives.list_ import make_list

                    return types.List.make(
                        {
                            "object_type": types.UnionType.make(
                                {
                                    "members": make_list(
                                        n_0=types.NoneType.make(), n_1=inner_res
                                    )
                                }
                            )
                        }
                    )

                inner_input_types = copy.copy(input_types)
                inner_input_types[mapped_param_name] = replacement
                inner_output_type = orig_op.output_type(inner_input_types)

                object_type = input_types[mapped_param_name].object_type

                if types.is_optional(object_type):
                    return types.List(types.optional(inner_output_type))
                return types.List(inner_output_type)

            output_type = make_output_type

        def resolve(**inputs):
            new_inputs = copy.copy(inputs)
            first_arg_name = list(new_inputs)[0]
            list_ = new_inputs.pop(first_arg_name)

            if op_policy.should_run_in_parallel(orig_op.name):
                # TODO: This whole resolver should be replaced with this
                # execution path. It is now the correct way to do it. But
                # I didn't update that path to handle / fix list tags, or to
                # return the right thing when this is a type resolver. So
                # we need to make those changes. Just doing this for now to
                # get new ops parallelized correctly.
                map_fn = weave_internal.define_fn(
                    {"row": first_arg.type}, lambda x: orig_op(x, **new_inputs)
                ).val
                res = execute_fast.fast_map_fn(list_, map_fn)
                return res

            tag_store_mem_map = tag_store._OBJ_TAGS_MEM_MAP.get()
            tag_store_curr_node_id = tag_store._OBJ_TAGS_CURR_NODE_ID.get()

            # Just do file-table in parallel for now. We'll do parallelization
            # more generally in the future.
            if (
                orig_op.name.endswith("file-table")
                or orig_op.name.endswith("file-joinedTable")
                or orig_op.name.endswith("file-partitionedTable")
                or orig_op.name.endswith("run-history")
                or orig_op.name.endswith("run-history2")
                or orig_op.name.endswith("run-history3")
            ):

                def download_one(x):
                    with tag_store.with_tag_store_state(
                        tag_store_curr_node_id, tag_store_mem_map
                    ):
                        if x == None or types.is_optional(first_arg.type):
                            return None
                        called = orig_op(x, **new_inputs)
                        # Use the use path to get caching.
                        try:
                            res = weave_internal.use(called)
                        except errors.WeaveArtifactCollectionNotFound:
                            return None
                        res = storage.deref(res)
                        return res

                if USE_PARALLEL_DOWNLOAD:
                    res = list(parallelism.do_in_parallel(download_one, list_))
                else:
                    res = list(map(download_one, list_))

                # file-table needs to flow tags, but we lose them going across the
                # thread boundary.
                for x, res_val in zip(list_, res):
                    if tag_store.is_tagged(x):
                        tag_store.add_tags(box.box(res_val), tag_store.get_tags(x))

                return res

            if isinstance(orig_op.concrete_output_type, types.TypeType):

                def refine_one(x) -> types.Type:
                    with tag_store.with_tag_store_state(
                        tag_store_curr_node_id, tag_store_mem_map
                    ):
                        if x != None or types.is_optional(first_arg.type):
                            return orig_op.resolve_fn(
                                **{mapped_param_name: x, **new_inputs}
                            )
                        return types.NoneType()

                if not list_:
                    return types.List(types.NoneType())
                if USE_PARALLEL_REFINE and orig_op.name in PARALLEL_REFINE_ALLOWLIST:
                    types_to_merge = list(parallelism.do_in_parallel(refine_one, list_))
                else:
                    types_to_merge = list(map(refine_one, list_))

                return types.List(types.union(*types_to_merge))

            # TODO: use the vectorization described here:
            # https://paper.dropbox.com/doc/Weave-Python-Weave0-Op-compatibility-workstream-kJ3XSDdgR96XwKPapHwPD
            list_tags = None
            if tag_store.is_tagged(list_) and orig_op._gets_tag_by_name != None:
                list_tags = tag_store.get_tags(list_)

            def resolve_one(x):
                with tag_store.with_tag_store_state(
                    tag_store_curr_node_id, tag_store_mem_map
                ):
                    if not (
                        x is None or isinstance(x, box.BoxedNone)
                    ) or types.is_optional(first_arg.type):
                        return orig_op.resolve_fn(
                            **{
                                mapped_param_name: tag_store.add_tags(
                                    box.box(x), list_tags
                                )
                                if list_tags is not None
                                else x,
                                **new_inputs,
                            }
                        )
                    return None

            if USE_PARALLEL_RESOLVE and orig_op.name in PARALLEL_RESOLVE_ALLOWLIST:
                return list(parallelism.do_in_parallel(resolve_one, list_))
            else:
                return list(map(resolve_one, list_))

        # Use the function signature of the original op to compute the signature
        # of the lazy call
        resolve.sig = inspect.signature(orig_op.raw_resolve_fn)  # type: ignore
        input_type = copy.copy(orig_op.input_type.arg_types)
        input_type[mapped_param_name] = types.List(types.optional(first_arg.type))
        new_op = op_def.OpDef(
            mapped_op_name,
            op_args.OpNamedArgs(input_type),
            output_type,
            resolve,
            _mapped_refine_output_type(orig_op),
        )

        # Note: The following commented out code block was written to
        # create a `weave_op` for every mapped op. This is then used
        # in vectorize as a faster fallback if an arrow equivalent does
        # not exist. However, before this writing, that code block was
        # not getting called / exercised. This is buggy in some way
        # and I could not get it working. Commenting out for now b/c I think
        # it is a good optimization, but not worth debugging right now.
        # def weave_fn_body(list_, *args):
        #     def map_item(item):
        #         full_named_args = {mapped_param_name: item}
        #         for i, na in enumerate(named_args[1:]):
        #             full_named_args[na.name] = args[i]

        #         # use Any type for OutputNode
        #         return graph.OutputNode(types.Any(), orig_op.name, full_named_args)

        #     return list_.map(lambda item: map_item(item))

        # new_op.weave_fn = weave_internal.define_fn(input_type, weave_fn_body).val
        op_version = registry_mem.memory_registry.register_op(new_op)

        return op_version

    @staticmethod
    def handle_class_decorator_update(
        derived_op: op_def.OpDef,
        base_weave_type: type[types.Type],
        orig_op_new_name: str,
    ):
        with context_state.legacy_strict_op_saving(False):
            named_args = derived_op.input_type.named_args()
            if len(named_args) == 0 or not isinstance(
                derived_op.input_type, op_args.OpNamedArgs
            ):
                raise errors.WeaveDefinitionError(
                    f"Expected mapped op {derived_op.name} to have named first argument."
                )
            first_arg = named_args[0]
            # Check to see if the first argument is a list of UnknownType. This is how
            # we know that the type is expected to be the class type
            first_arg_is_cls = first_arg.type == types.List(
                types.optional(types.UnknownType())
            )
            if first_arg_is_cls:
                derived_op.input_type.arg_types[first_arg.name] = types.List(
                    types.optional(base_weave_type())
                )
            registry_mem.memory_registry.rename_op(
                derived_op.name, MappedDeriveOpHandler.derived_name(orig_op_new_name)
            )


def _mapped_refine_output_type(orig_op):
    refine_output_type = None

    mapped_refine_op = (
        orig_op.refine_output_type.derived_ops["mapped"]
        if (
            orig_op.refine_output_type is not None
            and "mapped" in orig_op.refine_output_type.derived_ops
        )
        else None
    )

    if mapped_refine_op:

        def mapped_refine_output_type_refiner(*args, **kwargs):
            return mapped_refine_op.raw_resolve_fn(*args, **kwargs)
            # union_members = mapped_refine_op.raw_resolve_fn(*args, **kwargs)
            # if not union_members:
            #     return types.List(types.NoneType())
            # return types.List(
            #     types.union(
            #         *[types.NoneType() if mem == None else mem for mem in union_members]
            #     )
            # )

        unioned_mapped_type_refiner_op = op_def.OpDef(
            f"{mapped_refine_op.name}_unioned",
            mapped_refine_op.input_type,
            types.TypeType(),
            mapped_refine_output_type_refiner,
            None,
            plugins=orig_op.refine_output_type.plugins,
        )
        unioned_mapped_type_refiner_op_def = registry_mem.memory_registry.register_op(
            unioned_mapped_type_refiner_op
        )
        refine_output_type = unioned_mapped_type_refiner_op_def
    return refine_output_type


def handler_for_id(handler_id: str) -> type[DeriveOpHandler]:
    for handler in DeriveOpHandler.__subclasses__():
        if handler.handler_id == handler_id:
            return handler
    raise errors.WeaveInternalError(f"Unknown derive op handler {handler_id}")


def derive_ops(op: op_def.OpDef):
    with context_state.legacy_strict_op_saving(False):
        for handler in DeriveOpHandler.__subclasses__():
            if (
                handler.should_derive_op(op)
                and handler.handler_id not in op.derived_ops
            ):
                new_op = handler.make_derived_op(op)
                op.derived_ops[handler.handler_id] = new_op
                new_op.derived_from = op



from weave.legacy.weave import weave_types
from weave.legacy.weave import storage


class Client:
    def __init__(self, server):
        self.server = server

    def execute(self, nodes, no_cache=False):
        results = self.server.execute(nodes, no_cache=no_cache)

        # Deref if node output type is not RefType
        # TODO: move to language_ref.py, do in compile pass
        # TODO: this logic is duplicated in server.py:_handle_request
        return [
            r if isinstance(n.type, weave_types.RefType) else storage.deref(r)
            for (n, r) in zip(nodes, results)
        ]


class NonCachingClient:
    def __init__(self, server):
        self.server = server

    def execute(self, nodes):
        res = self.server.execute(nodes, no_cache=True)
        return [
            r if isinstance(n.type, weave_types.RefType) else storage.deref(r)
            for (n, r) in zip(nodes, res)
        ]



import datetime
import typing

import pytest

import weave
from weave.legacy.weave.ops_primitives import date


def assert_date_string(
    diff: typing.Union[float, int],
    unit: str,
    actual: str,
):
    if abs(diff) == 1:
        unit = unit[:-1]

    round_unit = 1 if unit == "years" or unit == "months" else 0
    diff = round(diff, round_unit)
    if int(diff) == diff:
        diff = int(diff)

    expected = f"{diff} {unit}"
    assert actual == expected


@pytest.mark.parametrize(
    "multiplier,unit,num_ms",
    [
        (m, u, nms)
        for m in [1, 1.5, 3, 5.3432, 8.00000000007]
        for u, nms in date.AUTO_FORMAT_UNITS_AND_NUM_MS
    ],
)
def test_relative_string_autoformat(multiplier, unit, num_ms):
    ts1 = datetime.datetime.now()
    ts2 = ts1 + datetime.timedelta(milliseconds=num_ms * multiplier)

    negative_result = weave.use(date.auto_format_relative_string(ts1, ts2))
    positive_result = weave.use(date.auto_format_relative_string(ts2, ts1))

    assert_date_string(multiplier, unit, positive_result)
    assert_date_string(-1 * multiplier, unit, negative_result)


@pytest.mark.parametrize("diff_ms", [0, 0.5, -0.5])
def test_relative_string_autoformat_edge_cases(diff_ms):
    ts1 = datetime.datetime.now()
    ts2 = ts1 + datetime.timedelta(milliseconds=diff_ms)

    result = weave.use(date.auto_format_relative_string(ts1, ts2))
    assert result == "less than 1 ms"



"""The top-level functions for Weave Trace API."""

import contextlib
import os
import threading
import time
from typing import Any, Iterator, Optional, Union

# TODO: type_serializers is imported here to trigger registration of the image serializer.
# There is probably a better place for this, but including here for now to get the fix in.
from weave import type_serializers  # noqa: F401
from weave.legacy.weave import urls
from weave.trace import util
from weave.trace.call_context import get_current_call, require_current_call
from weave.trace.client_context import weave_client as weave_client_context

from . import context, weave_client, weave_init
from .constants import TRACE_OBJECT_EMOJI
from .op import as_op, op
from .refs import ObjectRef, parse_uri
from .settings import UserSettings, parse_and_apply_settings, should_disable_weave
from .table import Table


def init(
    project_name: str,
    *,
    settings: Optional[Union[UserSettings, dict[str, Any]]] = None,
) -> weave_client.WeaveClient:
    """Initialize weave tracking, logging to a wandb project.

    Logging is initialized globally, so you do not need to keep a reference
    to the return value of init.

    Following init, calls of weave.op() decorated functions will be logged
    to the specified project.

    Args:
        project_name: The name of the Weights & Biases project to log to.

    Returns:
        A Weave client.
    """
    parse_and_apply_settings(settings)

    if should_disable_weave():
        return weave_init.init_weave_disabled().client

    return weave_init.init_weave(project_name).client


@contextlib.contextmanager
def remote_client(project_name: str) -> Iterator[weave_init.weave_client.WeaveClient]:
    inited_client = weave_init.init_weave(project_name)
    try:
        yield inited_client.client
    finally:
        inited_client.reset()


# This is currently an internal interface. We'll expose something like it though ("offline" mode)
def init_local_client() -> weave_client.WeaveClient:
    return weave_init.init_local().client


@contextlib.contextmanager
def local_client() -> Iterator[weave_client.WeaveClient]:
    inited_client = weave_init.init_local()
    try:
        yield inited_client.client
    finally:
        inited_client.reset()


def publish(obj: Any, name: Optional[str] = None) -> weave_client.ObjectRef:
    """Save and version a python object.

    If an object with name already exists, and the content hash of obj does
    not match the latest version of that object, a new version will be created.

    TODO: Need to document how name works with this change.

    Args:
        obj: The object to save and version.
        name: The name to save the object under.

    Returns:
        A weave Ref to the saved object.
    """
    client = weave_client_context.require_weave_client()

    save_name: str
    if name:
        save_name = name
    elif n := getattr(obj, "name", None):
        save_name = n
    else:
        save_name = obj.__class__.__name__

    ref = client._save_object(obj, save_name, "latest")

    if isinstance(ref, weave_client.ObjectRef):
        if isinstance(ref, weave_client.OpRef):
            url = urls.op_version_path(
                ref.entity,
                ref.project,
                ref.name,
                ref.digest,
            )
        # TODO(gst): once frontend has direct dataset/model links
        # elif isinstance(obj, weave_client.Dataset):
        else:
            url = urls.object_version_path(
                ref.entity,
                ref.project,
                ref.name,
                ref.digest,
            )
        print(f"{TRACE_OBJECT_EMOJI} Published to {url}")
    return ref


def ref(location: str) -> weave_client.ObjectRef:
    """Construct a Ref to a Weave object.

    TODO: what happens if obj does not exist

    Args:
        location: A fully-qualified weave ref URI, or if weave.init() has been called, "name:version" or just "name" ("latest" will be used for version in this case).


    Returns:
        A weave Ref to the object.
    """
    if not "://" in location:
        client = weave_client_context.get_weave_client()
        if not client:
            raise ValueError("Call weave.init() first, or pass a fully qualified uri")
        if "/" in location:
            raise ValueError("'/' not currently supported in short-form URI")
        if ":" not in location:
            name = location
            version = "latest"
        else:
            name, version = location.split(":")
        location = str(client._ref_uri(name, version, "obj"))

    uri = parse_uri(location)
    if not isinstance(uri, weave_client.ObjectRef):
        raise ValueError("Expected an object ref")
    return uri


def obj_ref(obj: Any) -> Optional[weave_client.ObjectRef]:
    return weave_client.get_ref(obj)


def output_of(obj: Any) -> Optional[weave_client.Call]:
    client = weave_client_context.require_weave_client()

    ref = obj_ref(obj)
    if ref is None:
        return ref

    return client._ref_output_of(ref)


@contextlib.contextmanager
def attributes(attributes: dict[str, Any]) -> Iterator:
    """
    Context manager for setting attributes on a call.

    Example:

    ```python
    with weave.attributes({'env': 'production'}):
        print(my_function.call("World"))
    ```
    """
    cur_attributes = {**context.call_attributes.get()}
    cur_attributes.update(attributes)

    token = context.call_attributes.set(cur_attributes)
    try:
        yield
    finally:
        context.call_attributes.reset(token)


def serve(
    model_ref: ObjectRef,
    method_name: Optional[str] = None,
    auth_entity: Optional[str] = None,
    port: int = 9996,
    thread: bool = False,
) -> str:
    import uvicorn

    from weave.legacy.weave import wandb_api

    from .serve_fastapi import object_method_app

    client = weave_client_context.require_weave_client()
    # if not isinstance(
    #     client, _graph_client_wandb_art_st.GraphClientWandbArtStreamTable
    # ):
    #     raise ValueError("serve currently only supports wandb client")

    print(f"Serving {model_ref}")
    print(f"🥐 Server docs and playground at http://localhost:{port}/docs")
    print()
    os.environ["PROJECT_NAME"] = f"{client.entity}/{client.project}"
    os.environ["MODEL_REF"] = str(model_ref)

    wandb_api_ctx = wandb_api.get_wandb_api_context()
    app = object_method_app(model_ref, method_name=method_name, auth_entity=auth_entity)
    trace_attrs = context.call_attributes.get()

    def run() -> None:
        # This function doesn't return, because uvicorn.run does not return.
        with wandb_api.wandb_api_context(wandb_api_ctx):
            with attributes(trace_attrs):
                uvicorn.run(app, host="0.0.0.0", port=port)

    if util.is_notebook():
        thread = True
    if thread:
        t = threading.Thread(target=run, daemon=True)
        t.start()
        time.sleep(1)
        return "http://localhost:%d" % port
    else:
        # Run should never return
        run()
    raise ValueError("Should not reach here")


def finish() -> None:
    """Stops logging to weave.

    Following finish, calls of weave.op() decorated functions will no longer be logged. You will need to run weave.init() again to resume logging.

    """
    weave_init.finish()


# As of this writing, most important symbols are
# re-exported in __init__.py.
# __docspec__ = []


__all__ = [
    "init",
    "remote_client",
    "local_client",
    "init_local_client",
    "as_op",
    "publish",
    "ref",
    "obj_ref",
    "output_of",
    "attributes",
    "serve",
    "finish",
    "op",
    "Table",
    "ObjectRef",
    "parse_uri",
    "get_current_call",
    "weave_client_context",
    "require_current_call",
]



import contextlib
import contextvars
import dataclasses
import enum
import logging
import logging.config
import os
import pathlib
import typing
import warnings
from logging.handlers import WatchedFileHandler

from pythonjsonlogger import jsonlogger

from weave.legacy.weave import environment

try:
    from flask.logging import wsgi_errors_stream
except ModuleNotFoundError:
    wsgi_errors_stream = None  # type: ignore


class LogFormat(enum.Enum):
    PRETTY = "pretty"
    DATADOG = "datadog"
    JSON = "json"


@dataclasses.dataclass
class LogSettings:
    format: LogFormat
    level: typing.Optional[int]


_log_indent: contextvars.ContextVar[int] = contextvars.ContextVar(
    "_log_indent", default=0
)


def increment_indent() -> contextvars.Token[int]:
    indent = _log_indent.get()
    return _log_indent.set(indent + 1)


def reset_indent(token: contextvars.Token[int]) -> None:
    _log_indent.reset(token)


@contextlib.contextmanager
def indent_logs() -> typing.Iterator:
    token = increment_indent()
    try:
        yield
    finally:
        reset_indent(token)


def get_indent() -> int:
    return _log_indent.get()


class IndentFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        indent = get_indent()
        record.indent = "  " * indent  # type: ignore
        return True


default_log_format = "[%(asctime)s] %(levelname)s in %(module)s (Thread Name: %(threadName)s): %(indent)s%(message)s"


def get_logdir() -> typing.Optional[str]:
    logdir = "/tmp/weave/log"
    try:
        pathlib.Path(logdir).mkdir(exist_ok=True, parents=True)
    except OSError:
        warnings.warn(
            f"weave: Unable make log dir '{logdir}'. Filesystem logging will be disabled for "
            f"the remainder of this session. To enable filesystem logging, ensure the path is writable "
            f"and restart the server."
        )
        return None
    return logdir


def get_logfile_path(logfile_path: str) -> typing.Optional[str]:
    logdir = get_logdir()
    if logdir is None:
        return None
    full_path = pathlib.Path(logdir) / logfile_path
    try:
        full_path.touch()
    except OSError:
        warnings.warn(
            f"weave: Unable to touch logfile at '{full_path}'. Filesystem logging will be disabled for "
            f"the remainder of this session. To enable filesystem logging, ensure the path is writable "
            f"and restart the server."
        )
        return None
    return str(full_path)


def default_log_filename() -> typing.Optional[str]:
    pid = os.getpid()
    return get_logfile_path(f"{pid}.log")


def env_log_level() -> int:
    level_str_from_env = os.environ.get("WEAVE_LOG_LEVEL")
    if level_str_from_env:
        level_int_from_env = logging.getLevelName(level_str_from_env.upper())
        if isinstance(level_int_from_env, int):
            return level_int_from_env
        print(
            f'WEAVE_LOG_LEVEL environment variable value "{level_str_from_env}" is invalid.'
        )

    if os.environ.get("WEAVE_SERVER_ENABLE_LOGGING"):
        return logging.INFO

    # Default to only showing ERROR logs
    return logging.ERROR


def silence_mpl() -> None:
    mpl_logger = logging.getLogger("matplotlib")
    if mpl_logger:
        mpl_logger.setLevel(logging.CRITICAL)


def set_global_log_level(level: int) -> None:
    # Unused, but runs through all existing loggers and sets their level to the given level
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    for logger_name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)


def print_handlers(logger: logging.Logger) -> None:
    # For debugging
    print(f"Handlers for logger '{logger.name}':")
    for handler in logger.handlers:
        print(f"  {handler}")


def print_all_handlers() -> None:
    # For debugging
    root_logger = logging.getLogger()
    print_handlers(root_logger)

    for logger_name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(logger_name)
        print_handlers(logger)


class WeaveJSONEncoder(jsonlogger.JsonEncoder):
    def default(self, obj: typing.Any) -> typing.Any:
        if obj is None:
            # This is needed because datadog strips keys with null values from logs
            return "<<_WEAVE_NONE_>>"
        return super().default(obj)  # type: ignore[no-untyped-call]


def setup_handler(handler: logging.Handler, settings: LogSettings) -> None:
    formatter = logging.Formatter(default_log_format)
    if settings.format == LogFormat.DATADOG:
        formatter = jsonlogger.JsonFormatter(
            "%(levelname)s [%(name)s] [%(filename)s:%(lineno)d] "
            "[dd.trace_id=%(dd.trace_id)s dd.span_id=%(dd.span_id)s] %(message)s",
            timestamp=True,
            json_encoder=WeaveJSONEncoder,
        )  # type: ignore[no-untyped-call]
    elif settings.format == LogFormat.JSON:
        formatter = jsonlogger.JsonFormatter(
            "%(levelname)s [%(name)s] [%(filename)s:%(lineno)d] %(message)s",
            timestamp=True,
            json_encoder=WeaveJSONEncoder,
        )  # type: ignore[no-untyped-call]
    handler.addFilter(IndentFilter())
    handler.setFormatter(formatter)
    if settings.level is not None:
        handler.setLevel(settings.level)


_LOGGING_CONFIGURED = False


def enable_stream_logging(
    logger: typing.Optional[logging.Logger] = None,
    wsgi_stream_settings: typing.Optional[LogSettings] = None,
    pid_logfile_settings: typing.Optional[LogSettings] = None,
    server_logfile_settings: typing.Optional[LogSettings] = None,
) -> None:
    global _LOGGING_CONFIGURED
    _LOGGING_CONFIGURED = True

    handler: logging.Handler
    logger = logger or logging.getLogger("root")

    if wsgi_errors_stream and wsgi_stream_settings is not None:
        handler = logging.StreamHandler(wsgi_errors_stream)  # type: ignore
        setup_handler(handler, wsgi_stream_settings)
        logger.addHandler(handler)

    if pid_logfile_settings is not None:
        log_filename = default_log_filename()
        if log_filename:
            handler = WatchedFileHandler(log_filename, mode="w")
            setup_handler(handler, pid_logfile_settings)
            logger.addHandler(handler)

    if server_logfile_settings is not None:
        log_filename = get_logfile_path("server.log")
        if log_filename:
            handler = logging.FileHandler(log_filename, mode="w")
            setup_handler(handler, server_logfile_settings)
            logger.addHandler(handler)


def configure_logger() -> None:
    if _LOGGING_CONFIGURED:
        return
    # Disable ddtrace logs, not sure why they're turned on.
    ddtrace_logs = logging.getLogger("ddtrace")
    ddtrace_logs.setLevel(logging.ERROR)

    logger = logging.getLogger("root")
    logger.setLevel(env_log_level())

    # Remove StreamHandler from the root logger to avoid stdout logging
    for handler in logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            logger.removeHandler(handler)

    dd_env = os.getenv("DD_ENV")
    enable_datadog = dd_env
    if not enable_datadog:
        # This is the standard path for users.
        if os.getenv("WEAVE_SERVER_ENABLE_LOGGING"):
            # WEAVE_SERVER_ENABLE_LOGGING forces the logs to go to the wsgi
            # stream which will go to the console. We set this flag in our
            # server start scripts, so that if you run the server in its own
            # terminal, you get the logs.
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(
                    environment.weave_log_format(LogFormat.PRETTY), level=None
                ),
                pid_logfile_settings=LogSettings(
                    environment.weave_log_format(LogFormat.PRETTY), logging.INFO
                ),
            )
        else:
            # This is the default case. Log to a file for this process, but
            # do not write to standard out/error. This is important because
            # when you run Weave in a notebook, it'll create a server, but
            # we don't want the logs to show up in the notebook.
            enable_stream_logging(
                logger,
                pid_logfile_settings=LogSettings(
                    environment.weave_log_format(LogFormat.PRETTY), logging.INFO
                ),
            )
    else:
        if dd_env == "ci":
            # CI expects logs in the pid logfile
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(
                    environment.weave_log_format(LogFormat.PRETTY), logging.INFO
                ),
                pid_logfile_settings=LogSettings(
                    environment.weave_log_format(LogFormat.PRETTY), logging.INFO
                ),
            )
        elif environment.wandb_production():
            # Only log in the datadog format to the wsgi stream
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(
                    environment.weave_log_format(LogFormat.DATADOG), level=None
                ),
            )
        else:
            # Otherwise this is dev mode with datadog logging turned on.
            # Log to standard out and a known filename that the datadog
            # agent can watch.
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(
                    environment.weave_log_format(LogFormat.PRETTY), level=None
                ),
                server_logfile_settings=LogSettings(
                    environment.weave_log_format(LogFormat.DATADOG), level=None
                ),
            )



import dataclasses
import typing

import weave
from weave.legacy.weave import graph, panel

ExpressionType = typing.TypeVar("ExpressionType")


@weave.type()
class FunctionEditorConfig(typing.Generic[ExpressionType]):
    expr: graph.Node[ExpressionType] = dataclasses.field(default_factory=graph.VoidNode)


@weave.type()
class FunctionEditor(panel.Panel):
    id = "FunctionEditor"
    config: FunctionEditorConfig = dataclasses.field(
        default_factory=FunctionEditorConfig
    )



# Ideas for ops, but not production ready.

from weave.legacy.weave import api as api
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, weave_class

op(
    name="root-compare_versions",
    input_type={"one_version": types.Any()},
    output_type=types.List(
        types.TypedDict(
            # Hardcode output type for Compare demo right now.
            {"x": types.Float(), "y": types.Float(), "version": types.String()}
        )
    ),
)


def compare_versions(one_version):
    versions = api.versions(one_version)
    data = []
    for i, version in enumerate(versions):
        v_data = version.get()
        label = str(api.expr(version))
        for row in v_data:
            row.update({"version": label})
            data.append(row)
    return data



import os
import typing

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave import weave_internal as internal
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import util, dispatch
from weave.legacy.weave.panels import panel_board, panel_group
from weave.legacy.weave.panels_py import panel_autoboard
from weave.legacy.weave.panels_py.generator_templates import template_registry

panels = weave.legacy.weave.panels
ops = weave.legacy.weave.ops


# BOARD_ID must be unique across all ops. It must only contain letters and underscores.
BOARD_ID = "llm_completions_monitor"

# BOARD_DISPLAY_NAME is the name that will be displayed in the UI
BOARD_DISPLAY_NAME = "Monitor OpenAI API usage"

# BOARD_DESCRIPTION is the description that will be displayed in the UI
BOARD_DESCRIPTION = "Use the OpenAI integration to track, monitor and analyze API calls. Understand performance and quality. Track trends and organization-wide costs."

# BOARD_INPUT_WEAVE_TYPE is the weave type of the input node.
BOARD_INPUT_WEAVE_TYPE = types.List(
    types.TypedDict(
        {
            "inputs": types.optional(
                types.TypedDict(
                    {
                        "messages": types.optional(
                            types.List(
                                types.TypedDict(
                                    {
                                        "role": types.optional(types.String()),
                                        "content": types.optional(types.String()),
                                    },
                                    not_required_keys=set(["role"]),
                                )
                            )
                        )
                    }
                )
            ),
            "output": types.optional(
                types.TypedDict(
                    {
                        "id": types.optional(types.String()),
                        "object": types.optional(types.String()),
                        "model": types.optional(types.String()),
                        "choices": types.optional(
                            types.List(
                                types.TypedDict(
                                    {
                                        "message": types.optional(
                                            types.TypedDict(
                                                {
                                                    "role": types.optional(
                                                        types.String()
                                                    ),
                                                    "content": types.optional(
                                                        types.String()
                                                    ),
                                                },
                                                not_required_keys=set(["role"]),
                                            )
                                        ),
                                        "finish_reason": types.optional(types.String()),
                                    },
                                    not_required_keys=set(["finish_reason"]),
                                )
                            )
                        ),
                    },
                    not_required_keys=set(["id", "object"]),
                )
            ),
            "timestamp": types.optional(types.Timestamp()),
            "summary": types.optional(
                types.TypedDict(
                    {
                        "prompt_tokens": types.optional(types.Number()),
                        "completion_tokens": types.optional(types.Number()),
                        "total_tokens": types.optional(types.Number()),
                    }
                )
            ),
        }
    )
)


board_name = "py_board-" + BOARD_ID


# @weave.type()
# class OpenAITable(weave.Panel):
#     id = "OpenAITable"
#     input_node: weave.Node[list[typing.Any]]

#     @weave.op()
#     def render(self) -> weave.legacy.weave.panels.Table:
#         table = panels.Table(
#             self.input_node,
#             columns=[lambda row: row["inputs"]["messages"][-1]["content"]],
#         )
#         # table.add_column( lambda row: row["inputs"]["messages"][-1]["content"], "Messages"
#         # )
#         # table.add_column(lambda row: row["messages"][-1]["content"], "Message")
#         return table

API_COST_INPUT_TYPE = types.TypedDict(
    {
        "output": types.optional(
            types.TypedDict(
                {
                    "model": types.optional(types.String()),
                }
            )
        ),
        "summary": types.optional(
            types.TypedDict(
                {
                    "prompt_tokens": types.optional(types.Number()),
                    "completion_tokens": types.optional(types.Number()),
                }
            )
        ),
    }
)


@weave.op(weavify=True, input_type={"record": API_COST_INPUT_TYPE})  # type: ignore
def openai_request_cost(record) -> float:  # type: ignore
    model = record["output.model"]
    pt = record["summary.prompt_tokens"]
    ct = record["summary.completion_tokens"]
    cost_per_1000 = weave.legacy.weave.ops.case(
        [
            # finetuned
            {"when": model.startsWith("ada:"), "then": pt * 0.0016 + ct * 0.0016},
            {"when": model.startsWith("babbage:"), "then": pt * 0.0024 + ct * 0.0024},
            {"when": model.startsWith("curie:"), "then": pt * 0.012 + ct * 0.012},
            {"when": model.startsWith("davinci:"), "then": pt * 0.12 + ct * 0.12},
            # non-finetuned
            {"when": model == "gpt-4-32k-0314", "then": pt * 0.06 + ct * 0.12},
            {"when": model == "gpt-4-32k-0613", "then": pt * 0.06 + ct * 0.12},
            {"when": model == "gpt-3.5-turbo-0613", "then": pt * 0.0015 + ct * 0.002},
            {
                "when": model == "gpt-3.5-turbo-16k-0613",
                "then": pt * 0.003 + ct * 0.004,
            },
            {
                "when": model == "text-embedding-ada-002-v2",
                "then": pt * 0.0001 + ct * 0.0001,
            },
            {"when": model == "ada", "then": pt * 0.0004 + ct * 0.0004},
            {"when": model == "babbage", "then": pt * 0.0005 + ct * 0.0005},
            {"when": model == "curie", "then": pt * 0.002 + ct * 0.002},
            {"when": model == "davinci", "then": pt * 0.02 + ct * 0.02},
            {
                "when": model.startsWith("gpt-3.5-turbo"),
                "then": pt * 0.002 + ct * 0.002,
            },
            {"when": model.startsWith("gpt-4"), "then": pt * 0.03 + ct * 0.06},
            {"when": True, "then": 0},
        ]
    )
    return cost_per_1000 / 1000  # type: ignore


@weave.op(  # type: ignore
    name=board_name,
    hidden=True,
    input_type={
        "input_node": types.Function(
            {},
            BOARD_INPUT_WEAVE_TYPE,
        )
    },
)
def board(
    input_node,
) -> panels.Group:
    timestamp_col_name = "timestamp"

    ### Varbar

    # Add the input node as raw data
    varbar = panel_board.varbar(editable=False)

    source_data = varbar.add("source_data", input_node)

    augmented_data = varbar.add(
        "augmented_data",
        source_data.with_columns(
            weave.legacy.weave.ops.dict_(
                **{
                    "summary.cost": source_data.map(
                        lambda row: openai_request_cost(row)
                    )
                }
            )
        ),
        hidden=True,
    )

    filter_fn = varbar.add(
        "filter_fn",
        weave_internal.define_fn(
            {"row": input_node.type.object_type}, lambda row: weave_internal.const(True)
        ),
        hidden=True,
    )

    grouping_fn = varbar.add(
        "grouping_fn",
        weave_internal.define_fn(
            {"row": input_node.type.object_type}, lambda row: row["output.model"]
        ),
        hidden=True,
    )

    filtered_data = varbar.add(
        "filtered_data", augmented_data.filter(filter_fn), hidden=True
    )

    # Setup date range variables:
    ## 1. raw_data_range is derived from raw_data
    filtered_range = varbar.add(
        "filtered_range",
        weave.legacy.weave.ops.make_list(
            a=filtered_data[timestamp_col_name].min(),
            b=filtered_data[timestamp_col_name].max(),
        ),
        hidden=True,
    )

    ## 2. user_zoom_range is used to store the user's zoom range
    user_zoom_range = varbar.add("user_zoom_range", None, hidden=True)

    ## 2.b: Setup a date picker to set the user_zoom_range
    varbar.add(
        "time_range",
        weave.legacy.weave.panels.DateRange(
            user_zoom_range, domain=source_data[timestamp_col_name]
        ),
    )

    ## 3. bin_range is derived from user_zoom_range and raw_data_range. This is
    ##    the range of data that will be displayed in the charts.
    bin_range = varbar.add(
        "bin_range", user_zoom_range.coalesce(filtered_range), hidden=True
    )
    # Derive the windowed data to use in the plots as a function of bin_range

    window_data = varbar.add(
        "window_data",
        augmented_data.filter(
            lambda row: weave.legacy.weave.ops.Boolean.bool_and(
                row[timestamp_col_name] >= bin_range[0],
                row[timestamp_col_name] <= bin_range[1],
            )
        ),
        hidden=True,
    )

    filters = varbar.add(
        "filters",
        weave.legacy.weave.panels.FilterEditor(filter_fn, node=window_data),
    )

    filtered_window_data = varbar.add(
        "filtered_window_data", window_data.filter(filter_fn), hidden=True
    )

    grouping = varbar.add(
        "grouping",
        weave.legacy.weave.panels.GroupingEditor(grouping_fn, node=window_data),
    )

    height = 5

    ### Overview tab

    overview_tab = weave.legacy.weave.panels.Group(
        layoutMode="grid",
        showExpressions=True,
        enableAddPanel=True,
    )  # , showExpressions="titleBar")
    overview_tab.add(
        "request_count",
        panel_autoboard.timeseries(
            filtered_data,
            bin_domain_node=bin_range,
            x_axis_key="timestamp",
            y_expr=lambda row: row.count(),
            y_title="request count",
            color_expr=lambda row: grouping_fn(row),
            color_title="group",
            x_domain=user_zoom_range,
            n_bins=100,
            mark="bar",
        ),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=0, w=24, h=height),
    )

    overview_tab.add(
        "cost",
        panel_autoboard.timeseries(
            filtered_data,
            bin_domain_node=bin_range,
            x_axis_key="timestamp",
            y_expr=lambda row: row["summary.cost"].sum(),
            y_title="total cost ($)",
            color_expr=lambda row: grouping_fn(row),
            color_title="group",
            x_domain=user_zoom_range,
            n_bins=50,
        ),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=height, w=12, h=height),
    )

    overview_tab.add(
        "latency",
        panel_autoboard.timeseries(
            filtered_data,
            bin_domain_node=bin_range,
            x_axis_key="timestamp",
            y_expr=lambda row: row["summary.latency_s"].avg(),
            y_title="avg latency (s)",
            color_expr=lambda row: grouping_fn(row),
            color_title="group",
            x_domain=user_zoom_range,
            n_bins=50,
        ),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=12, y=height, w=12, h=height),
    )

    overview_tab.add(
        "avg_cost_per_req",
        filtered_window_data["summary.cost"].avg(),  # type: ignore
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=height * 2, w=6, h=3),
    )
    overview_tab.add(
        "avg_prompt_tokens_per_req",
        filtered_window_data["summary.prompt_tokens"].avg(),  # type: ignore
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=6, y=height * 2, w=6, h=3),
    )
    overview_tab.add(
        "avg_completion_tokens_per_req",
        filtered_window_data["summary.completion_tokens"].avg(),  # type: ignore
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=12, y=height * 2, w=6, h=3),
    )
    (
        overview_tab.add(
            "avg_total_tokens_per_req",
            filtered_window_data["summary.total_tokens"].avg(),  # type: ignore
            layout=weave.legacy.weave.panels.GroupPanelLayout(x=18, y=height * 2, w=6, h=3),
        ),
    )

    # Show a plot for each attribute.
    # TODO: This doesn't really work yet (needs some manual UI configuration currently,
    # and it's ugly).
    # overview_tab.add(
    #     "attributes", weave.legacy.weave.panels.EachColumn(filtered_window_data["attributes"])
    # )

    ### Requests tab

    # requests_tab = weave.legacy.weave.panels.Group(
    #     layoutMode="grid",
    #     showExpressions=True,
    # )  # l, showExpressions="titleBar")

    requests_table = panels.Table(filtered_window_data)  # type: ignore
    requests_table.add_column(lambda row: row["output.model"], "Model")
    requests_table.add_column(
        lambda row: row["inputs.messages"][-1]["content"], "Last Prompt"
    )
    requests_table.add_column(
        lambda row: row["output.choices"][-1]["message.content"],
        "Completion",
    )
    requests_table.add_column(lambda row: row["summary.prompt_tokens"], "Prompt Tokens")
    requests_table.add_column(
        lambda row: row["summary.completion_tokens"], "Completion Tokens"
    )
    requests_table.add_column(lambda row: row["summary.total_tokens"], "Total Tokens")
    requests_table.add_column(lambda row: row["summary.latency_s"], "Latency")
    requests_table.add_column(lambda row: row["summary.cost"], "Cost")
    requests_table.add_column(
        lambda row: row["timestamp"], "Timestamp", sort_dir="desc"
    )

    requests_table_var = overview_tab.add(
        "table",
        requests_table,
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=13, w=24, h=8),
    )
    overview_tab.add(
        "input",
        panels.Table(  # type: ignore
            requests_table_var.active_data()["inputs.messages"],
            columns=[lambda row: row["role"], lambda row: row["content"]],
        ),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=21, w=12, h=8),
    )
    overview_tab.add(
        "output",
        requests_table_var.active_row(),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=12, y=21, w=12, h=8),
    )

    # attributes_tab = weave.legacy.weave.panels.Group(layoutMode="grid")

    # users_tab = weave.legacy.weave.panels.Group(layoutMode="grid")

    # models_tab = weave.legacy.weave.panels.Group(layoutMode="grid")

    # tabs = panels.Group(
    #     layoutMode="tab",
    #     items={
    #         "Overview": overview_tab,
    #         "Requests": requests_tab,
    #         # "Attributes": attributes_tab,
    #         # "Users": users_tab,
    #         # "Models": models_tab,
    #     },
    # )

    return panels.Board(vars=varbar, panels=overview_tab)


instructions_md = util.read_or_default(
    os.path.join(os.path.dirname(__file__), "instructions", "panel_llm_monitor.md")
)

template_registry.register(
    board_name,
    BOARD_DISPLAY_NAME,
    BOARD_DESCRIPTION,
    is_featured=True,
    instructions_md=instructions_md,
    thumbnail_url="https://raw.githubusercontent.com/wandb/weave/master/docs/assets/monitor-open-ai-api-usage.png",
)



from weave.legacy.weave import context_state as _context

_loading_builtins_token = _context.set_loading_built_ins()
from weave.legacy.weave.ecosystem.keras.model import *

_context.clear_loading_built_ins(_loading_builtins_token)



from typing import List, Optional

from openai.types.chat import (
    ChatCompletion,
    ChatCompletionMessage,
    ChatCompletionMessageParam,
)
from pydantic import BaseModel, Field

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.monitoring import monitor


class ModelTokensConfig(BaseModel):
    per_message: int = 0
    per_name: int = 0


class CombinedChoice(BaseModel):
    content: str = ""
    finish_reason: Optional[str] = None
    role: Optional[str] = None
    function_call: Optional[str] = None
    tool_calls: Optional[str] = None


class ChatCompletionRequest(BaseModel):
    class Config:
        use_enum_values = True

    model: str = ""
    messages: List[ChatCompletionMessageParam] = Field(default_factory=list)
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    stream: Optional[bool] = None


class Context(BaseModel):
    class Config:
        extra = "allow"

    # span: Optional[monitor.Span]
    inputs: Optional[ChatCompletionRequest] = None
    outputs: Optional[ChatCompletion] = None



import contextlib
import dataclasses
import functools
import logging
import os
import re
import tempfile
import typing
from urllib import parse

import requests
from wandb import Artifact
from wandb.apis.public import api as wb_public
from wandb.sdk.lib.hashutil import b64_to_hex_id, hex_to_b64_id

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import (
    filesystem,
    urls,
    errors,
    engine_trace,
    artifact_fs,
    eager,
    file_base,
    file_util,
    memo,
    uris,
    wandb_client_api,
    util,
)
from weave.legacy.weave.wandb_interface import wandb_artifact_pusher

if typing.TYPE_CHECKING:
    from weave.legacy.weave.wandb_interface.wandb_lite_run import InMemoryLazyLiteRun

    from .run_streamtable_span import RunStreamTableSpan


quote_slashes = functools.partial(parse.quote, safe="")

DEFAULT_WEAVE_OBJ_PROJECT = "weave"


class WandbArtifactManifestEntry(typing.TypedDict):
    digest: str
    birthArtifactID: typing.Optional[str]
    ref: typing.Optional[str]
    size: typing.Optional[int]
    extra: typing.Optional[dict[str, typing.Any]]
    # wandb.sdk.wandb_artifacts.ArtifactManifestEntry also has "local_path"
    # but that doesn't exist in saved manifests


class WandbArtifactManifestV1(typing.TypedDict):
    version: int
    storagePolicy: str
    storagePolicyConfig: typing.Optional[dict[str, typing.Any]]
    contents: typing.Dict[str, WandbArtifactManifestEntry]


# We used to use wandb.sdk.wandb_artifacts.ArtifactManifest directly, but it
# is expensive to construct because it makes new objects for every entry and
# causes a lot of memory/gc churn. This implementation leaves the manifest
# as the raw json dict instead.
@dataclasses.dataclass
class WandbArtifactManifest:
    class StorageLayout:
        V1 = "V1"
        V2 = "V2"

    _manifest_json: WandbArtifactManifestV1

    @property
    def _storage_policy_config(self):
        return self._manifest_json.get("storagePolicyConfig", {})

    @property
    def storage_layout(self):
        return self._storage_policy_config.get(
            "storageLayout", WandbArtifactManifest.StorageLayout.V1
        )

    def get_entry_by_path(
        self, path: str
    ) -> typing.Optional[WandbArtifactManifestEntry]:
        return self._manifest_json["contents"].get(path)

    def get_paths_in_directory(self, path: str) -> typing.List[str]:
        if path == "":
            return list(self._manifest_json["contents"].keys())
        dir_path = path + "/"
        return [
            k for k in self._manifest_json["contents"].keys() if k.startswith(dir_path)
        ]


# TODO: Get rid of this, we have the new wandb api service! But this
# is still used in a couple places.
@memo.memo  # Per-request memo reduces duplicate calls to the API
def get_wandb_read_artifact(path: str):
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_artifact"):
        try:
            return wandb_client_api.wandb_public_api().artifact(path)
        except wandb_client_api.WandbCommError:
            raise errors.WeaveArtifactVersionNotFound(
                f"Could not find artifact with path {path} in W&B"
            )


def is_valid_version_index(version_index: str) -> bool:
    return bool(re.match(r"^v(0|[1-9][0-9]*)$", version_index))


def get_wandb_read_artifact_uri(path: str):
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_artifact_uri"):
        art = get_wandb_read_artifact(path)
        return WeaveWBArtifactURI(
            art.name.split(":", 1)[0],
            art.commit_hash,
            art.entity,
            art.project,
        )


def wandb_artifact_dir():
    d = os.path.join(filesystem.get_filesystem_dir(), "wandb_artifacts")
    os.makedirs(d, exist_ok=True)
    return d


@dataclasses.dataclass
class ReadClientArtifactURIResult:
    weave_art_uri: "WeaveWBArtifactURI"
    artifact_type_name: str
    is_deleted: bool


def _art_id_is_client_version_id_mapping(art_id: str) -> bool:
    return len(art_id) == 128 and ":" not in art_id


def _art_id_is_client_collection_and_alias_id_mapping(art_id: str) -> bool:
    return ":" in art_id


def _convert_client_id_to_server_id(art_id: str) -> str:
    query = wb_public.gql(
        """
        query ClientIDMapping($clientID: ID!) {
            clientIDMapping(clientID: $clientID) {
                serverID
            }
        }
    """
    )
    tracer = engine_trace.tracer()
    with tracer.trace("_convert_client_id_to_server_id.execute"):
        res = wandb_client_api.wandb_public_api().client.execute(
            query,
            variable_values={
                "clientID": art_id,
            },
        )
    return b64_to_hex_id(res["clientIDMapping"]["serverID"])


def _collection_and_alias_id_mapping_to_uri(
    client_collection_id: str, alias_name: str
) -> ReadClientArtifactURIResult:
    is_deleted = False
    query = """
    query ArtifactVersionFromIdAlias(
        $id: ID!,
        $aliasName: String!
    ) {
        artifactCollection(id: $id) {
            id
            name
            state
            project {
                id
                name
                entity {
                    id
                    name
                }
            }
            artifactMembership(aliasName: $aliasName) {
                id
                versionIndex
                commitHash
            }
            artifactMemberships(first: 1) {
                edges {
                    node {
                        id
                        versionIndex
                        commitHash
                    }
                }
            }
            defaultArtifactType {
                id
                name
            }
        }
    }
    """

    try:
        tracer = engine_trace.tracer()
        with tracer.trace("_collection_and_alias_id_mapping_to_uri.execute"):
            res = wandb_client_api.query_with_retry(
                query,
                variables={
                    "id": client_collection_id,
                    "aliasName": alias_name,
                },
                num_timeout_retries=1,
            )
    except requests.exceptions.HTTPError as e:
        if e.response is not None and e.response.status_code == 400:
            # This is a special case: the client id corresponds to an artifact that was
            # never uploaded, so the client id doesn't exist in the W&B server.

            collection = None
            logging.warn(
                f"Artifact collection with client id {client_collection_id} not present in W&B server."
            )

        else:
            raise e
    else:
        collection = res["artifactCollection"]

    if collection is None:
        # Note: deleted collections are still returned by the API (with state=DELETED)
        # So a missing collection is a real error (unless it was never uploaded)
        raise errors.WeaveArtifactCollectionNotFound(
            f"Could not find artifact collection with client id {client_collection_id}"
        )
    elif collection["state"] != "READY":  # state is either "DELETED" or "READY"
        # This is a deleted artifact, but we still have a record of it.
        is_deleted = True

    artifact_type_name = res["artifactCollection"]["defaultArtifactType"]["name"]
    artifact_membership = res["artifactCollection"]["artifactMembership"]
    if artifact_membership is None:
        # Here we account for a special case: when the alias is `latest`. It is
        # exceedingly common and expected to reference the latest alias of a
        # collection. This is particularly common when logging tables (but used
        # in many other places as well). Occasionally the user deletes the
        # `latest` alias which is a super easy thing to do in the UI. However,
        # this results in the data not being found. We solved this similarly in
        # the typescript implementation by getting the most recent version in
        # such cases. At one point we had an extended discussion about making
        # `latest` a "calculated" alias, which would solve this issue on
        # gorilla's side, but were unable to reach an actionable conclusion.
        # This is a simple workaround that solves for these cases.
        if (
            alias_name == "latest"
            and len(res["artifactCollection"]["artifactMemberships"]["edges"]) > 0
        ):
            commit_hash = res["artifactCollection"]["artifactMemberships"]["edges"][0][
                "node"
            ]["commitHash"]
        else:
            is_deleted = True
            # If the membership is deleted, then we will not be able to get the commit hash.
            # Here, we can simply use the alias name as the commit hash. If the collection
            # is ever un-deleted or the alias is re-assigned, then the next call will result
            # in a "true" commit hash, ensuring we don't hit a stale cache.
            commit_hash = alias_name
    else:
        commit_hash = artifact_membership["commitHash"]
    entity_name = collection["project"]["entity"]["name"]
    project_name = collection["project"]["name"]
    artifact_name = collection["name"]

    weave_art_uri = WeaveWBArtifactURI(
        artifact_name,
        commit_hash,
        entity_name,
        project_name,
    )

    return ReadClientArtifactURIResult(weave_art_uri, artifact_type_name, is_deleted)


def _version_server_id_to_uri(server_id: str) -> ReadClientArtifactURIResult:
    is_deleted = False
    query = wb_public.gql(
        """
    query ArtifactVersionFromServerId(
        $id: ID!,
    ) {
        artifact(id: $id) {
            id
            state
            commitHash
            versionIndex
            artifactType {
                id
                name
            }
            artifactSequence {
                id
                name
                state
                project {
                    id
                    name
                    entity {
                        id
                        name
                    }
                }
            }
        }
    }
    """
    )
    tracer = engine_trace.tracer()
    with tracer.trace("_version_server_id_to_uri.execute"):
        res = wandb_client_api.wandb_public_api().client.execute(
            query,
            variable_values={"id": hex_to_b64_id(server_id)},
        )

    artifact = res["artifact"]
    if artifact is None:
        # Note: deleted versions are still returned by the API (with state=DELETED)
        # So a missing artifact is a real error.
        raise errors.WeaveArtifactVersionNotFound(
            f"Could not find artifact version with server id {server_id}"
        )
    elif (
        artifact["state"] != "COMMITTED"
    ):  # state is either "DELETED" or "PENDING" or "COMMITTED"
        # This is a deleted artifact, but we still have a record of it
        is_deleted = True

    collection = artifact["artifactSequence"]

    if collection is None:
        # Note: deleted collections are still returned by the API (with state=DELETED)
        # So a missing collection is a real error.
        raise errors.WeaveArtifactCollectionNotFound(
            f"Could not find artifact sequence for artifact version with server id {server_id}"
        )
    elif collection["state"] != "READY":  # state is either "DELETED" or "READY"
        # This is a deleted artifact, but we still have a record of it
        is_deleted = True

    artifact_type_name = artifact["artifactType"]["name"]
    commit_hash = artifact["commitHash"]

    entity_name = collection["project"]["entity"]["name"]
    project_name = collection["project"]["name"]
    artifact_name = collection["name"]

    weave_art_uri = WeaveWBArtifactURI(
        artifact_name,
        commit_hash,
        entity_name,
        project_name,
    )

    return ReadClientArtifactURIResult(weave_art_uri, artifact_type_name, is_deleted)


def get_wandb_read_client_artifact_uri(art_id: str) -> ReadClientArtifactURIResult:
    """art_id may be client_id, seq:alias, or server_id"""
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_client_artifact_uri"):
        if _art_id_is_client_version_id_mapping(art_id):
            server_id = _convert_client_id_to_server_id(art_id)
            return _version_server_id_to_uri(server_id)
        elif _art_id_is_client_collection_and_alias_id_mapping(art_id):
            client_collection_id, alias_name = art_id.split(":")
            return _collection_and_alias_id_mapping_to_uri(
                client_collection_id, alias_name
            )
        else:
            return _version_server_id_to_uri(art_id)


def get_wandb_read_client_artifact(art_id: str) -> typing.Optional["WandbArtifact"]:
    """art_id may be client_id, seq:alias, or server_id"""
    tracer = engine_trace.tracer()
    with tracer.trace("get_wandb_read_client_artifact"):
        res = get_wandb_read_client_artifact_uri(art_id)
        if res.is_deleted:
            return None
        return WandbArtifact(
            res.weave_art_uri.name, res.artifact_type_name, res.weave_art_uri
        )


class WandbArtifactType(artifact_fs.FilesystemArtifactType):
    def save_instance(self, obj, artifact, name) -> "WandbArtifactRef":
        return WandbArtifactRef(obj, None)


class WandbArtifact(artifact_fs.FilesystemArtifact):
    def __init__(
        self,
        name,
        type=None,
        uri: typing.Optional[
            typing.Union[
                "WeaveWBArtifactURI",
                "WeaveWBLoggedArtifactURI",
                "WeaveWBArtifactByIDURI",
            ]
        ] = None,
    ):
        from weave.legacy.weave import io_service

        self.io_service = io_service.get_sync_client()
        self.name = name

        # original uri passed, if any. this can include aliases such as latest or best, which do not
        # correspond to specific versions or may change versions.
        self._unresolved_read_artifact_or_client_uri = None

        # resolved version of the URI above. this points to the same artifact as the unresolved URI but
        # includes a specific, immutable version like v4 instead of an alias. this is needed for idempotency
        # of cache keys.
        self._resolved_read_artifact_uri: typing.Optional[
            typing.Union["WeaveWBArtifactURI", "WeaveWBArtifactByIDURI"]
        ] = None
        self._read_artifact = None
        if not uri:
            self._writeable_artifact = Artifact(
                name, "op_def" if type is None else type
            )
        else:
            # load an existing artifact, this should be read only,
            # TODO: we could technically support writable artifacts by creating a new version?
            self._unresolved_read_artifact_or_client_uri = uri
        self._local_path: dict[str, str] = {}
        # self._path_handlers: dict[str, typing.Any] = {}

    @property
    def branch(self) -> typing.Optional[str]:
        if self._unresolved_read_artifact_or_client_uri is not None:
            branch = self._unresolved_read_artifact_or_client_uri.version

        if branch is not None and not likely_commit_hash(branch):
            return branch

        return None

    @property
    def branch_point(self) -> typing.Optional[artifact_fs.BranchPointType]:
        # all branching is local for now. we could support remote branches
        # but in this initial implementation, assume only LocalArtifacts
        # live on branches.
        return None

    @property
    def _read_artifact_uri(
        self,
    ) -> typing.Optional[typing.Union["WeaveWBArtifactURI", "WeaveWBArtifactByIDURI"]]:
        if self._resolved_read_artifact_uri is not None:
            return self._resolved_read_artifact_uri

        if isinstance(
            self._unresolved_read_artifact_or_client_uri,
            (WeaveWBLoggedArtifactURI, WeaveWBArtifactURI, WeaveWBArtifactByIDURI),
        ):
            self._resolved_read_artifact_uri = (
                self._unresolved_read_artifact_or_client_uri.resolved_artifact_uri
            )
        return self._resolved_read_artifact_uri

    @property
    def _ref(self) -> "WandbArtifactRef":
        # existing_ref = ref_base.get_ref(obj)
        # if isinstance(existing_ref, artifact_base.ArtifactRef):
        #     return existing_ref
        if not self.is_saved:
            raise errors.WeaveInternalError("cannot get ref of an unsaved artifact")
        return WandbArtifactRef(self, None, None)

    def _set_read_artifact_uri(self, uri):
        self._read_artifact = None
        self._resolved_read_artifact_uri = None
        self._unresolved_read_artifact_or_client_uri = uri

    # TODO: still using wandb lib for this, but we should switch to the new
    # wandb api service
    @property
    def _saved_artifact(self):
        if self._read_artifact is None:
            uri = self._read_artifact_uri
            if isinstance(uri, WeaveWBArtifactByIDURI):
                raise errors.WeaveInternalError(
                    "The `_saved_artifact` property does not support `wandb-artifact-by-id` URI scheme. This is an internal programming error"
                )
            path = uri.uri_path
            self._read_artifact = get_wandb_read_artifact(path)
        return self._read_artifact

    def __repr__(self):
        return "<WandbArtifact %s>" % self.name

    def delete(self) -> None:
        self._saved_artifact.delete(delete_aliases=True)

    def add_artifact_reference(self, uri: str) -> None:
        # The URI for a get node might look like: wandb-artifact:///<entity>/<project>/test8:latest/obj
        # but the SDK add_reference call requires something like:
        # wandb-artifact://41727469666163743a353432353830303535/obj.object.json
        if uri.startswith("wandb-artifact-by-id:///"):
            raise errors.WeaveInternalError(
                "`add_artifact_reference` does not support `wandb-artifact-by-id` URI scheme. This is an internal programming error"
            )
        assert uri.startswith("wandb-artifact:///")
        uri_parts = WeaveWBArtifactURI.parse(uri)
        uri_path = uri_parts.uri_path
        ref_artifact = get_wandb_read_artifact(uri_path)

        # A reference needs to point to a specific existing file in the other artifact.
        # For stream tables obj.object.json should exist. For logged tables, the filename
        # can vary, so we just pick the first file we find.
        filename = "obj.object.json"
        try:
            ref_url = ref_artifact.get_path(filename).ref_url()
        except KeyError:
            for ref_file in ref_artifact.files(None, per_page=1):
                ref_url = ref_artifact.get_path(ref_file.name).ref_url()
                break

        # We have a name collision problem. E.g. A board and a stream table will both have
        # a file named obj.object.json. Add a prefix to distinguish them.
        ref_name = f"{ref_artifact.id}__{filename}"
        self._writeable_artifact.add_reference(ref_url, ref_name)

    @property
    def commit_hash(self) -> str:
        if not self.is_saved:
            raise errors.WeaveInternalError(
                "cannot get commit hash of an unsaved artifact"
            )
        resolved_uri = self._read_artifact_uri
        if resolved_uri is None:
            raise errors.WeaveInternalError("cannot resolve commit hash of artifact")
        version = resolved_uri.version
        if version is None:
            raise errors.WeaveInternalError("resolved uri has no version")
        return version

    @property
    def is_saved(self) -> bool:
        return (
            self._unresolved_read_artifact_or_client_uri is not None
            or self._read_artifact is not None
        )

    @property
    def version(self):
        if not self.is_saved:
            raise errors.WeaveInternalError("cannot get version of an unsaved artifact")

        # Note: W&B Artifact's `version` is often the `v0` alias, which is not what
        # we want here (despite the same name for the property). In Weave world, `version`
        # refers to either: a) a commit hash, or b) a branch name. Therefore, in this
        # case, we explicitly want the commit hash, not the differently-defined `version`.
        return self._saved_artifact.commit_hash

    @property
    def created_at(self):
        raise NotImplementedError()

    def get_other_version(self, version):
        raise NotImplementedError()

    def direct_url(self, path: str) -> typing.Optional[str]:
        if self._read_artifact_uri is None:
            raise errors.WeaveInternalError(
                'cannot get direct url for unsaved artifact"'
            )
        uri = self._read_artifact_uri.with_path(path)
        return self.io_service.direct_url(uri)

    def path(self, name: str) -> str:
        if not self.is_saved or not self._read_artifact_uri:
            raise errors.WeaveInternalError("cannot download of an unsaved artifact")

        uri = self._read_artifact_uri.with_path(name)
        fs_path = self.io_service.ensure_file(uri)
        if fs_path is None:
            # Important to raise FileNotFoundError here, FileSystemArtifactRef.type
            # relies on this.
            raise FileNotFoundError("Path not in artifact %s %s" % (self, name))
        return self.io_service.fs.path(fs_path)

    def size(self, path: str) -> int:
        manifest = self._manifest()
        if manifest is not None:
            manifest_entry = manifest.get_entry_by_path(path)
            if manifest_entry:
                size = manifest_entry["size"]
                if size is None:
                    return 0
                return size

        if path in self._saved_artifact.manifest.entries:
            return self._saved_artifact.manifest.entries[path].size
        return super().size(path)

    @property
    def initial_uri_obj(
        self,
    ) -> typing.Union[
        "WeaveWBArtifactURI", "WeaveWBLoggedArtifactURI", "WeaveWBArtifactByIDURI"
    ]:
        if not self.is_saved or not self._unresolved_read_artifact_or_client_uri:
            raise errors.WeaveInternalError("cannot get uri of an unsaved artifact")
        return self._unresolved_read_artifact_or_client_uri

    @property
    def uri_obj(self) -> typing.Union["WeaveWBArtifactURI", "WeaveWBArtifactByIDURI"]:
        if not self.is_saved or not self._read_artifact_uri:
            raise errors.WeaveInternalError("cannot get uri of an unsaved artifact")
        return self._read_artifact_uri

    @contextlib.contextmanager
    def new_file(self, path, binary=False):
        if not self._writeable_artifact:
            raise errors.WeaveInternalError("cannot add new file to readonly artifact")
        mode = "w"
        if binary:
            mode = "wb"
        with self._writeable_artifact.new_file(path, mode) as f:
            yield f

    @contextlib.contextmanager
    def writeable_file_path(self, path):
        with tempfile.TemporaryDirectory() as tmpdir:
            full_path = os.path.join(tmpdir, path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            yield full_path
            self._writeable_artifact.add_file(full_path, path)

    @contextlib.contextmanager
    def new_dir(self, path):
        if not self._writeable_artifact:
            raise errors.WeaveInternalError("cannot add new file to readonly artifact")
        with tempfile.TemporaryDirectory() as tmpdir:
            yield os.path.abspath(tmpdir)
            self._writeable_artifact.add_dir(tmpdir, path)

    @contextlib.contextmanager
    def open(self, path, binary=False):
        if not self.is_saved:
            raise errors.WeaveInternalError("cannot load data from an unsaved artifact")
        mode = "r"
        if binary:
            mode = "rb"
        p = self.path(path)
        with file_util.safe_open(p, mode) as f:
            yield f

    # def get_path_handler(self, path, handler_constructor):
    #     handler = self._path_handlers.get(path)
    #     if handler is None:
    #         handler = handler_constructor(self, path)
    #         self._path_handlers[path] = handler
    #     return handler

    def read_metadata(self):
        raise NotImplementedError()

    def write_metadata(self, dirname):
        raise NotImplementedError()

    def save(
        self,
        project: str = DEFAULT_WEAVE_OBJ_PROJECT,
        entity_name: typing.Optional[str] = None,
        branch: typing.Optional[str] = None,
        artifact_collection_exists: bool = False,
        *,
        _lite_run: typing.Optional["InMemoryLazyLiteRun"] = None,
    ):
        # for handler in self._path_handlers.values():
        #     handler.close()
        additional_aliases = [] if branch is None else [branch]
        res = wandb_artifact_pusher.write_artifact_to_wandb(
            self._writeable_artifact,
            project,
            entity_name,
            additional_aliases,
            artifact_collection_exists=artifact_collection_exists,
            _lite_run=_lite_run,
        )
        version = res.version_str if branch is None else branch
        self._set_read_artifact_uri(
            WeaveWBArtifactURI(
                name=res.artifact_name,
                version=version,
                entity_name=res.entity_name,
                project_name=res.project_name,
            )
        )

    def _manifest(self) -> typing.Optional[WandbArtifactManifest]:
        if self._read_artifact_uri is None:
            raise errors.WeaveInternalError(
                'cannot get path info for unsaved artifact"'
            )
        return self.io_service.manifest(self._read_artifact_uri)

    def digest(self, path: str) -> typing.Optional[str]:
        manifest_entry = self._manifest_entry(path)
        if manifest_entry is not None:
            return manifest_entry["digest"]
        return None

    def _manifest_entry(self, path: str) -> typing.Optional[WandbArtifactManifestEntry]:
        manifest = self._manifest()
        if manifest is not None:
            manifest_entry = manifest.get_entry_by_path(path)
            if manifest_entry is not None:
                return manifest_entry
        return None

    def _path_info(
        self, path: str
    ) -> typing.Optional[
        typing.Union[
            "artifact_fs.FilesystemArtifactFile",
            "artifact_fs.FilesystemArtifactDir",
            "artifact_fs.FilesystemArtifactRef",
        ]
    ]:
        manifest = self._manifest()
        if manifest is None:
            return None
        manifest_entry = manifest.get_entry_by_path(path)
        if manifest_entry is not None:
            # This is not a WeaveURI! Its the artifact reference style used
            # by the W&B Artifacts/media layer.
            ref_prefix = "wandb-artifact://"
            ref = manifest_entry.get("ref")
            if ref and ref.startswith(ref_prefix):
                # This is a reference to another artifact
                art_id, target_path = ref[len(ref_prefix) :].split("/", 1)
                art = get_wandb_read_client_artifact(art_id)
                # this should be None when the requested artifact is deleted from the server.
                # we want to return None in this case so that the caller can handle it.
                if art is None:
                    return None
                return artifact_fs.FilesystemArtifactFile(art, target_path)
            # This is a file
            return artifact_fs.FilesystemArtifactFile(self, path)

        # This is not a file, assume its a directory. If not, we'll return an empty result.
        dir_paths = manifest.get_paths_in_directory(path)
        sub_dirs: dict[str, file_base.SubDir] = {}
        files = {}
        total_size = 0
        sub_dir_sizes = {}
        cwd = os.getcwd()
        for entry_path in dir_paths:
            entry = manifest.get_entry_by_path(entry_path)
            path_size = (
                entry["size"] if entry is not None and entry["size"] is not None else 0
            )
            total_size += path_size

            # this used to be os.path.relpath but that called os.getcwd() every time
            # that turned out to be a bottleneck in production for artifacts with many
            # dir paths, so we use our own implementation that takes the cwd as input
            # and doesn't need to ever call os.getcwd()
            rel_path = util.relpath_no_syscalls(entry_path, path, cwd)
            rel_path_parts = rel_path.split("/")
            if len(rel_path_parts) == 1:
                files[rel_path_parts[0]] = artifact_fs.FilesystemArtifactFile(
                    self,
                    entry_path,
                )
            else:
                dir_name = rel_path_parts[0]
                if dir_name not in sub_dirs:
                    dir_ = file_base.SubDir(entry_path, 0, {}, {})
                    sub_dir_sizes[dir_name] = 0
                    sub_dirs[dir_name] = dir_
                sub_dir_sizes[dir_name] += path_size
                dir_ = sub_dirs[dir_name]
                if len(rel_path_parts) == 2:
                    dir_files = typing.cast(dict, dir_.files)
                    dir_files[rel_path_parts[1]] = artifact_fs.FilesystemArtifactFile(
                        self,
                        entry_path,
                    )
                else:
                    dir_.dirs[rel_path_parts[1]] = 1
        if not sub_dirs and not files:
            return None
        for dir_name, dir_ in sub_dirs.items():
            dir_.size = sub_dir_sizes[dir_name]
        return artifact_fs.FilesystemArtifactDir(
            self, path, total_size, sub_dirs, files
        )

    def _get_file_paths(self) -> list[str]:
        manifest = self._manifest()
        if manifest is None:
            raise errors.WeaveInternalError("No manifest when fetching file paths")
        return manifest.get_paths_in_directory("")

    @property
    def metadata(self) -> artifact_fs.ArtifactMetadata:
        mutable_metadata = {}
        readonly_metadata = {}
        if (
            hasattr(self, "_writeable_artifact")
            and self._writeable_artifact is not None
        ):
            mutable_metadata = self._writeable_artifact.metadata
        if self.is_saved:
            readonly_metadata = self._saved_artifact.metadata
        return artifact_fs.ArtifactMetadata(mutable_metadata, readonly_metadata)


WandbArtifactType.instance_classes = WandbArtifact


class WandbArtifactRef(artifact_fs.FilesystemArtifactRef):
    artifact: WandbArtifact

    def versions(self) -> list[artifact_fs.FilesystemArtifactRef]:
        # TODO: implement versions on wandb artifact
        return [self]

    @classmethod
    def from_uri(cls, uri: uris.WeaveURI) -> "WandbArtifactRef":
        if not isinstance(
            uri, (WeaveWBArtifactURI, WeaveWBLoggedArtifactURI, WeaveWBArtifactByIDURI)
        ):
            raise errors.WeaveInternalError(
                f"Invalid URI class passed to WandbArtifactRef.from_uri: {type(uri)}"
            )
        # TODO: potentially need to pass full entity/project/name instead
        return cls(
            WandbArtifact(uri.name, uri=uri),
            path=uri.path,
            extra=uri.extra,
        )

    @property
    def ui_url(self):
        root_type = self.type.root_type_class()
        from weave.legacy.weave.op_def_type import OpDefType

        if issubclass(root_type, OpDefType):
            return urls.op_version_path(
                self.artifact.uri_obj.entity_name,
                self.artifact.uri_obj.project_name,
                self.artifact.uri_obj.name,
                self.artifact.uri_obj.version,
            )
        else:
            return urls.object_version_path(
                self.artifact.uri_obj.entity_name,
                self.artifact.uri_obj.project_name,
                self.artifact.uri_obj.name,
                self.artifact.uri_obj.version,
            )

        # Before Tim's Weaveflow changes
        # return f"http://localhost:3000/browse2/{self.artifact.uri_obj.entity_name}/{self.artifact.uri_obj.project_name}/{self.type.root_type_class().name}/{self.artifact.uri_obj.name}/{self.artifact.uri_obj.version}"


types.WandbArtifactRefType.instance_class = WandbArtifactRef
types.WandbArtifactRefType.instance_classes = WandbArtifactRef

WandbArtifact.RefClass = WandbArtifactRef


def is_hex_string(s: str) -> bool:
    try:
        int(s, 16)
        return True
    except ValueError:
        return False


WANDB_COMMIT_HASH_LENGTH = 20


def likely_commit_hash(version: str) -> bool:
    # This is the heuristic we use everywhere to tell if a version is an alias or not
    # if self.version:
    return is_hex_string(version) and len(version) == WANDB_COMMIT_HASH_LENGTH


# Used to refer to objects stored in WB Artifacts. This URI must not change and
# matches the existing artifact schemes
@dataclasses.dataclass
class WeaveWBArtifactURI(uris.WeaveURI):
    SCHEME = "wandb-artifact"
    entity_name: str
    project_name: str
    netloc: typing.Optional[str] = None
    path: typing.Optional[str] = None
    extra: typing.Optional[list[str]] = None

    # resolved version of the artifact uri, in which aliases are replaced with specific versions
    _resolved_artifact_uri: typing.Optional["WeaveWBArtifactURI"] = None

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        parts = path.strip("/").split("/")
        parts = [parse.unquote(part) for part in parts]
        if len(parts) < 3:
            raise errors.WeaveInvalidURIError(f"Invalid WB Artifact URI: {uri}")
        entity_name = parts[0]
        project_name = parts[1]

        if ":" in parts[2]:
            name, version = parts[2].split(":", 1)
        else:
            name = parts[2]
            version = None

        file_path: typing.Optional[str] = None
        if len(parts) > 3:
            file_path = "/".join(parts[3:])

        extra: typing.Optional[list[str]] = None
        if fragment:
            extra = fragment.split("/")
        return cls(
            name,
            version,
            entity_name,
            project_name,
            netloc,
            file_path,
            extra,
        )

    @classmethod
    def parse(cls: typing.Type["WeaveWBArtifactURI"], uri: str) -> "WeaveWBArtifactURI":
        return super().parse(uri)  # type: ignore

    def __str__(self) -> str:
        netloc = self.netloc or ""
        uri = (
            f"{self.SCHEME}://"
            f"{quote_slashes(netloc)}/"
            f"{quote_slashes(self.entity_name)}/"
            f"{quote_slashes(self.project_name)}/"
            f"{quote_slashes(self.name)}:{self.version if self.version else ''}"
        )
        if self.path:
            uri += f"/{quote_slashes(self.path)}"
        if self.extra:
            uri += f"#{'/'.join([quote_slashes(e) for e in self.extra])}"
        return uri

    def with_path(self, path: str) -> "WeaveWBArtifactURI":
        return WeaveWBArtifactURI(
            self.name,
            self.version,
            self.entity_name,
            self.project_name,
            self.netloc,
            path,
            self.extra,
        )

    @property
    def uri_path(self) -> str:
        return f"{self.entity_name}/{self.project_name}/{self.name}:{self.version}"

    @property
    def resolved_artifact_uri(self) -> "WeaveWBArtifactURI":
        if self.version and likely_commit_hash(self.version):
            return self
        if self._resolved_artifact_uri is None:
            path = f"{self.entity_name}/{self.project_name}/{self.name}"
            if self.version:
                path += f":{self.version}"
            resolved_artifact_uri = get_wandb_read_artifact_uri(path)
            self._resolved_artifact_uri = resolved_artifact_uri
        return self._resolved_artifact_uri

    def to_ref(self) -> WandbArtifactRef:
        return WandbArtifactRef.from_uri(self)


@dataclasses.dataclass
class WeaveWBLoggedArtifactURI(uris.WeaveURI):
    SCHEME = "wandb-logged-artifact"
    # wandb-logged-artifact://afdsjaksdjflkasjdf12341234hv12h3v4k12j3hv41kh4v1423k14v1k2j3hv1k2j3h4v1k23h4v:[version|latest]/path
    #  scheme                                 name                                                              version      path
    path: typing.Optional[str] = None
    extra: typing.Optional[list[str]] = None

    # private attrs

    # resolved version of the artifact uri, in which aliases are replaced with specific versions
    _resolved_artifact_uri: typing.Optional[WeaveWBArtifactURI] = None

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        path = parse.unquote(path.strip("/"))
        spl_netloc = netloc.split(":")
        if len(spl_netloc) == 1:
            name = spl_netloc[0]
            version = None
        elif len(spl_netloc) == 2:
            name, version = spl_netloc
        else:
            raise errors.WeaveInvalidURIError(f"Invalid WB Client Artifact URI: {uri}")

        return WeaveWBLoggedArtifactURI(
            name=name,
            version=version,
            path=path,
        )

    def __str__(self) -> str:
        netloc = self.name
        if self.version:
            netloc += f":{self.version}"
        path = self.path or ""
        if path != "":
            path = f"/{path}"
        return f"{self.SCHEME}://{netloc}{path}"

    @property
    def resolved_artifact_uri(self) -> WeaveWBArtifactURI:
        if self._resolved_artifact_uri is None:
            art_id = self.name
            if self.version:
                # version is the alias or version index string
                art_id += f":{self.version}"
            res = get_wandb_read_client_artifact_uri(art_id)
            self._resolved_artifact_uri = res.weave_art_uri
        return self._resolved_artifact_uri.with_path(self.path or "")

    @property
    def entity_name(self) -> str:
        return self.resolved_artifact_uri.entity_name

    @property
    def project_name(self) -> str:
        return self.resolved_artifact_uri.project_name

    def with_path(self, path: str) -> "WeaveWBLoggedArtifactURI":
        return WeaveWBLoggedArtifactURI(
            self.name,
            self.version,
            path,
        )

    @classmethod
    def parse(cls, uri: str) -> "WeaveWBLoggedArtifactURI":
        scheme, netloc, path, params, query_s, fragment = parse.urlparse(uri)
        return cls.from_parsed_uri(
            uri, scheme, netloc, path, params, parse.parse_qs(query_s), fragment
        )

    def to_ref(self) -> WandbArtifactRef:
        return WandbArtifactRef.from_uri(self)


# This URI class allows us to build an artifact's uri without being dependent on its
# entity or project name. It is specifically relevant for the org registries use case
# where a team's artifact can be 'published' to a registry for users from other teams
# to view and use.
# When an artifact is created with an id-uri, it allows us to query the artifact's manifest and
# files just using the artifactID.
@dataclasses.dataclass
class WeaveWBArtifactByIDURI(uris.WeaveURI):
    SCHEME = "wandb-artifact-by-id"
    path_root = "__wb_artifacts_by_id__"
    entity_name = None
    project_name = None
    artifact_id: str
    netloc: typing.Optional[str] = None
    path: typing.Optional[str] = None
    extra: typing.Optional[list[str]] = None

    # resolved version of the artifact uri, in which aliases are replaced with specific versions
    _resolved_artifact_uri: typing.Optional["WeaveWBArtifactByIDURI"] = None

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        parts = path.strip("/").split("/")
        parts = [parse.unquote(part) for part in parts]
        if len(parts) < 3:
            raise errors.WeaveInvalidURIError(f"Invalid WB Artifact URI: {uri}")
        artifact_id = parts[1]

        if ":" in parts[2]:
            name, version = parts[2].split(":", 1)
        else:
            name = parts[2]
            version = None

        file_path: typing.Optional[str] = None
        if len(parts) > 3:
            file_path = "/".join(parts[3:])

        extra: typing.Optional[list[str]] = None
        if fragment:
            extra = fragment.split("/")
        return cls(
            name,
            version,
            artifact_id,
            netloc,
            file_path,
            extra,
        )

    @classmethod
    def parse(
        cls: typing.Type["WeaveWBArtifactByIDURI"], uri: str
    ) -> "WeaveWBArtifactByIDURI":
        return super().parse(uri)  # type: ignore

    def __str__(self) -> str:
        netloc = self.netloc or ""
        uri = (
            f"{self.SCHEME}://"
            f"{quote_slashes(netloc)}/"
            f"{self.path_root}/"
            f"{quote_slashes(self.artifact_id)}/"
            f"{quote_slashes(self.name)}:{self.version if self.version else ''}"
        )
        if self.path:
            uri += f"/{quote_slashes(self.path)}"
        if self.extra:
            uri += f"#{'/'.join([quote_slashes(e) for e in self.extra])}"
        return uri

    def with_path(self, path: str) -> "WeaveWBArtifactByIDURI":
        return WeaveWBArtifactByIDURI(
            self.name,
            self.version,
            self.artifact_id,
            self.netloc,
            path,
            self.extra,
        )

    @property
    def resolved_artifact_uri(self) -> "WeaveWBArtifactByIDURI":
        if self.version and likely_commit_hash(self.version):
            return self
        if self._resolved_artifact_uri is None:
            path = f"{self.path_root}/{self.artifact_id}/{self.name}"
            if self.version:
                path += f":{self.version}"
            # Since this class is only used while querying for artifact files,
            # we can assume the version exists and directly return the resolved
            # artifact instead of querying the public api for it
            resolved_artifact_uri = WeaveWBArtifactByIDURI(
                self.name.split(":", 1)[0],
                self.version,
                self.artifact_id,
            )
            self._resolved_artifact_uri = resolved_artifact_uri
        return self._resolved_artifact_uri

    def to_ref(self) -> WandbArtifactRef:
        return WandbArtifactRef.from_uri(self)


# This is a wrapper around an artifact that acts like a list of files.
# It fetchs a file from the manifest on __getItem__ and can return a count without fetching all files
@dataclasses.dataclass
class FilesystemArtifactFileIterator(list[artifact_fs.FilesystemArtifactFile]):
    data: list[str]
    artifact: WandbArtifact
    idx: int = 0

    def __init__(self, artifact: WandbArtifact, data: list[str] = []):
        self.data = data if len(data) > 0 else artifact._get_file_paths()
        self.artifact = artifact
        self.idx = 0

    def __getitem__(self, key):
        path_or_paths = self.data[key]
        if isinstance(path_or_paths, str):
            return self.artifact._path_info(path_or_paths)
        elif isinstance(path_or_paths, list):
            return FilesystemArtifactFileIterator(self.artifact, path_or_paths)
        raise errors.WeaveInternalError(
            "Invalid key in FilesystemArtifactFileIterator __getItem__"
        )

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self):
        return self

    def __next__(self):
        if self.idx >= len(self.data):
            raise StopIteration
        current_element = self.__getitem__(self.idx)
        self.idx += 1
        return current_element



import collections
import contextlib
import contextvars
import typing

from weave.legacy.weave import graph

from weave.legacy.weave import errors

ExecutableNode = typing.Union[graph.OutputNode, graph.ConstNode]


class NoResult:
    pass


class ErrorResult:
    error: Exception

    def __init__(self, error: Exception) -> None:
        self.error = error


class NodeResultStore:
    _from_store: typing.Optional[collections.defaultdict[graph.OutputNode, typing.Any]]
    _store: collections.defaultdict[graph.OutputNode, typing.Any]

    def __init__(
        self, initialize_with: typing.Optional["NodeResultStore"] = None
    ) -> None:
        if initialize_with is None:
            self._from_store = None
        else:
            self._from_store = initialize_with._store
        self._store = collections.defaultdict(lambda: NoResult)

    def has(self, node: graph.OutputNode) -> bool:
        return not self[node] is NoResult

    def __setitem__(self, key: graph.OutputNode, value: typing.Any):
        self._store[key] = value

    def __getitem__(self, key: graph.OutputNode) -> typing.Any:
        res = self._store[key]
        if isinstance(res, NoResult) and self._from_store is not None:
            return self._from_store[key]
        return res

    def merge(self, other: "NodeResultStore"):
        for key, value in other._store.items():
            self._store[key] = value


_node_result_store: contextvars.ContextVar[
    typing.Optional[NodeResultStore]
] = contextvars.ContextVar("_top_level_forward_graph_ctx", default=None)


# Each top level call to execute.execute_nodes creates its own result store.
# Recursive calls share the top level result store.
@contextlib.contextmanager
def node_result_store(store: typing.Optional[NodeResultStore] = None):
    if store is None:
        store = _node_result_store.get()
    token = None
    if store is None:
        store = NodeResultStore()
        token = _node_result_store.set(store)
    try:
        yield store
    finally:
        if token is not None:
            _node_result_store.reset(token)


def get_node_result_store() -> NodeResultStore:
    store = _node_result_store.get()
    if store is None:
        return NodeResultStore()
    return store


class ForwardNode:
    node: graph.OutputNode[ExecutableNode]
    input_to: dict["ForwardNode", typing.Literal[True]]

    def __init__(self, node: graph.OutputNode[ExecutableNode]) -> None:
        self.node = node
        self.input_to = {}
        self.cache_id = None
        self.result_store = get_node_result_store()

    def __str__(self):
        return "<ForwardNode(%s): %s input_to %s>" % (
            id(self),
            self.node.from_op.name
            if isinstance(self.node, graph.OutputNode)
            else self.node.val,
            " ".join([str(id(fn)) for fn in self.input_to]),
        )

    @property
    def result(self) -> typing.Any:
        return self.result_store[self.node]

    @property
    def has_result(self) -> bool:
        return self.result_store.has(self.node)

    def set_result(self, result: typing.Any) -> None:
        self.result_store[self.node] = result


# Each execute.execute_nodes call gets its own ForwardGraph, and walks all
# the nodes in it. But it will skip executing nodes if the result already
# exist for that node in the top-level result store.
class ForwardGraph:
    roots: dict[ForwardNode, typing.Literal[True]]
    _node_to_forward_node: typing.Dict[graph.Node, ForwardNode]

    def __init__(self, allow_var_nodes=False) -> None:
        self.roots = {}
        self._node_to_forward_node = {}
        self._allow_var_nodes = allow_var_nodes

    def add_nodes(self, nodes: typing.Iterable[graph.Node]):
        for node in nodes:
            self.add_node(node)

    def add_node(self, node: graph.Node):
        if isinstance(node, graph.VoidNode):
            raise errors.WeaveBadRequest(
                "Found void node when constructing ForwardGraph: %s" % node
            )
        elif isinstance(node, graph.ConstNode):
            return
        if node in self._node_to_forward_node:
            return
        if isinstance(node, graph.VarNode) and not self._allow_var_nodes:
            raise errors.WeaveBadRequest(
                "Found var node when constructing ForwardGraph: %s" % node
            )

        forward_node = ForwardNode(node)  # type: ignore
        self._node_to_forward_node[node] = forward_node
        if isinstance(node, graph.OutputNode):
            is_root = True
            for param_node in node.from_op.inputs.values():
                self.add_node(param_node)
                if isinstance(param_node, graph.OutputNode):
                    self._node_to_forward_node[param_node].input_to[forward_node] = True
                    is_root = False
            if is_root:
                self.roots[forward_node] = True

    def get_forward_node(self, node: typing.Union[graph.OutputNode, graph.VarNode]):
        return self._node_to_forward_node[node]

    def has_result(self, node: ExecutableNode):
        if isinstance(node, graph.ConstNode):
            return True
        return self._node_to_forward_node[node].has_result

    def get_result(self, node: ExecutableNode):
        if isinstance(node, graph.ConstNode):
            return node.val
        return self._node_to_forward_node[node].result

    def __str__(self):
        lines = []
        to_print = self.roots
        printed = set()
        while len(to_print):
            printing_now = to_print.copy()
            to_print = set()
            for forward_node in printing_now:
                lines.append(str(forward_node))
                for downstream_forward_node in forward_node.input_to:
                    if downstream_forward_node not in printed:
                        to_print.add(downstream_forward_node)
                    printed.add(downstream_forward_node)
        return "\n".join(lines)



import typing

from weave.trace import autopatch, errors, init_message, trace_sentry, weave_client
from weave.trace.client_context import weave_client as weave_client_context
from weave.trace_server import sqlite_trace_server
from weave.trace_server_bindings import remote_http_trace_server

_current_inited_client = None


class InitializedClient:
    def __init__(self, client: weave_client.WeaveClient):
        self.client = client
        weave_client_context.set_weave_client_global(client)

    def reset(self) -> None:
        weave_client_context.set_weave_client_global(None)


def get_username() -> typing.Optional[str]:
    from weave.legacy.weave import wandb_api

    api = wandb_api.get_wandb_api_sync()
    try:
        return api.username()
    except AttributeError:
        return None


def get_entity_project_from_project_name(project_name: str) -> tuple[str, str]:
    from weave.legacy.weave import wandb_api

    fields = project_name.split("/")
    if len(fields) == 1:
        api = wandb_api.get_wandb_api_sync()
        entity_name = api.default_entity_name()
        if entity_name is None:
            raise errors.WeaveWandbAuthenticationException(
                'weave init requires wandb. Run "wandb login"'
            )
        project_name = fields[0]
    elif len(fields) == 2:
        entity_name, project_name = fields
    else:
        raise ValueError(
            'project_name must be of the form "<project_name>" or "<entity_name>/<project_name>"'
        )
    if not entity_name:
        raise ValueError("entity_name must be non-empty")

    return entity_name, project_name


"""
This is the main entrypoint for the weave library. It initializes the weave client
and sets up the global state for the weave library.

Args:
    project_name (str): The project name to use for the weave client.
    ensure_project_exists (bool): If True (default), the client will attempt to create the project if it does not exist.
"""


def init_weave(
    project_name: str, ensure_project_exists: bool = True
) -> InitializedClient:
    global _current_inited_client
    if _current_inited_client is not None:
        # TODO: Prob should move into settings
        if (
            _current_inited_client.client.project == project_name
            and _current_inited_client.client.ensure_project_exists
            == ensure_project_exists
        ):
            return _current_inited_client
        else:
            _current_inited_client.reset()

    from weave.legacy.weave import wandb_api

    # Must init to read ensure we've read auth from the environment, in
    # case we're on a new thread.
    wandb_api.init()
    wandb_context = wandb_api.get_wandb_api_context()
    if wandb_context is None:
        import wandb

        print("Please login to Weights & Biases (https://wandb.ai/) to continue:")
        wandb.login(anonymous="never", force=True)  # type: ignore
        wandb_api.init()
        wandb_context = wandb_api.get_wandb_api_context()

    entity_name, project_name = get_entity_project_from_project_name(project_name)
    wandb_run_id = weave_client.safe_current_wb_run_id()
    weave_client.check_wandb_run_matches(wandb_run_id, entity_name, project_name)

    api_key = None
    if wandb_context is not None and wandb_context.api_key is not None:
        api_key = wandb_context.api_key
    remote_server = init_weave_get_server(api_key)
    # from .trace_server.clickhouse_trace_server_batched import ClickHouseTraceServer

    # server = ClickHouseTraceServer(host="localhost")
    client = weave_client.WeaveClient(
        entity_name, project_name, remote_server, ensure_project_exists
    )
    # If the project name was formatted by init, update the project name
    project_name = client.project

    _current_inited_client = InitializedClient(client)
    # entity_name, project_name = get_entity_project_from_project_name(project_name)
    # from .trace_server.clickhouse_trace_server_batched import ClickHouseTraceServer

    # client = weave_client.WeaveClient(ClickHouseTraceServer(host="localhost"))

    # init_client = InitializedClient(client)  # type: ignore

    # autopatching is only supported for the wandb client, because OpenAI calls are not
    # logged in local mode currently. When that's fixed, this autopatch call can be
    # moved to InitializedClient.__init__
    autopatch.autopatch()

    username = get_username()
    try:
        min_required_version = (
            remote_server.server_info().min_required_weave_python_version
        )
    # TODO: Tighten this exception to only catch the specific exception
    # that is thrown by the server_info call.
    except Exception:
        # Set to a minimum version that will always pass the check
        # In the future, we may want to throw here.
        min_required_version = "0.0.0"
    init_message.assert_min_weave_version(min_required_version)
    init_message.print_init_message(
        username, entity_name, project_name, read_only=not ensure_project_exists
    )

    user_context = {"username": username} if username else None
    trace_sentry.global_trace_sentry.configure_scope(
        {
            "entity_name": entity_name,
            "project_name": project_name,
            "user": user_context,
        }
    )

    return _current_inited_client


def init_weave_disabled() -> InitializedClient:
    """Initialize a dummy client that does nothing.

    This is used when the program is execuring with Weave disabled.

    Note: as currently implemented, any explicit calls to client.{X} will
    likely fail, since the user is not authenticated. The purpose of
    disabling weave is to disable _tracing_. Programs that attempt to
    make requests (eg. publishing, fetching, querying) while disabled
    will fail.
    """
    global _current_inited_client
    if _current_inited_client is not None:
        _current_inited_client.reset()

    client = weave_client.WeaveClient(
        "DISABLED",
        "DISABLED",
        init_weave_get_server("DISABLED", should_batch=False),
        ensure_project_exists=False,
    )

    return InitializedClient(client)


def init_weave_get_server(
    api_key: typing.Optional[str] = None,
    should_batch: bool = True,
) -> remote_http_trace_server.RemoteHTTPTraceServer:
    res = remote_http_trace_server.RemoteHTTPTraceServer.from_env(should_batch)
    if api_key is not None:
        res.set_auth(("api", api_key))
    return res


def init_local() -> InitializedClient:
    server = sqlite_trace_server.SqliteTraceServer("weave.db")
    server.setup_tables()
    client = weave_client.WeaveClient("none", "none", server)
    return InitializedClient(client)


def finish() -> None:
    global _current_inited_client
    if _current_inited_client is not None:
        _current_inited_client.reset()
        _current_inited_client = None

    # autopatching is only supported for the wandb client, because OpenAI calls are not
    # logged in local mode currently. When that's fixed, this reset_autopatch call can be
    # moved to InitializedClient.reset
    autopatch.reset_autopatch()
    trace_sentry.global_trace_sentry.end_session()



"""
This module contains code for checking if there is a new version of a PyPI package available.

It is a modified version of the code from the wandb library, located at:
https://github.com/wandb/client/blob/main/wandb/sdk/internal/update.py

Copied here to avoid a dependency on the wandb library and allow more pointed control
over the version checking logic.
"""

import queue
import sys
import threading
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
    Union,
)

import requests

if TYPE_CHECKING:
    import packaging.version  # type: ignore[import-not-found]


def check_available(
    current_version: str, module_name: str
) -> Optional[Dict[str, Optional[str]]]:
    """
    Check if there is a new version of the module available on PyPI.

    Args:
        current_version (str): The current version of the module.
        module_name (str): The name of the module to check for updates.

    Returns:
        Optional[Dict[str, Optional[str]]]: A dictionary containing the upgrade message, yank message, or delete message, or None if no update is available.
    """
    package_info = _find_available(current_version, module_name)
    if not package_info:
        return None

    latest_version, pip_prerelease, deleted, yanked, yanked_reason = package_info
    upgrade_message = (
        "{} version {} is available!  To upgrade, please run:\n"
        " $ pip install {} --upgrade{}".format(
            module_name,
            latest_version,
            module_name,
            " --pre" if pip_prerelease else "",
        )
    )
    delete_message = None
    if deleted:
        delete_message = "{} version {} has been retired!  Please upgrade.".format(
            module_name,
            current_version,
        )
    yank_message = None
    if yanked:
        reason_message = "({})  ".format(yanked_reason) if yanked_reason else ""
        yank_message = "{} version {} has been recalled!  {}Please upgrade.".format(
            module_name,
            current_version,
            reason_message,
        )

    # A new version is available!
    return {
        "upgrade_message": upgrade_message,
        "yank_message": yank_message,
        "delete_message": delete_message,
    }


def _parse_version(version: str) -> "packaging.version.Version":
    """Parse a version string into a version object.

    This function is a wrapper around the `packaging.version.parse` function, which
    is used to parse version strings into version objects. If the `packaging` library
    is not installed, it falls back to the `pkg_resources` library.
    """
    try:
        from packaging.version import parse as parse_version  # type: ignore
    except ImportError:
        from pkg_resources import parse_version  # type: ignore[assignment]

    return parse_version(version)


def _async_call(
    target: Callable, timeout: Optional[Union[int, float]] = None
) -> Callable:
    """Wrap a method to run in the background with an optional timeout.

    Returns a new method that will call the original with any args, waiting for upto
    timeout seconds. This new method blocks on the original and returns the result or
    None if timeout was reached, along with the thread. You can check thread.is_alive()
    to determine if a timeout was reached. If an exception is thrown in the thread, we
    reraise it.
    """
    q: queue.Queue = queue.Queue()

    def wrapped_target(q: "queue.Queue", *args: Any, **kwargs: Any) -> Any:
        try:
            q.put(target(*args, **kwargs))
        except Exception as e:
            q.put(e)

    def wrapper(
        *args: Any, **kwargs: Any
    ) -> Union[Tuple[Exception, "threading.Thread"], Tuple[None, "threading.Thread"]]:
        thread = threading.Thread(
            target=wrapped_target, args=(q,) + args, kwargs=kwargs
        )
        thread.daemon = True
        thread.start()
        try:
            result = q.get(True, timeout)
            if isinstance(result, Exception):
                raise result.with_traceback(sys.exc_info()[2])
            return result, thread
        except queue.Empty:
            return None, thread

    return wrapper


def _find_available(
    current_version: str, module_name: str
) -> Optional[Tuple[str, bool, bool, bool, Optional[str]]]:
    pypi_url = f"https://pypi.org/pypi/{module_name}/json"
    yanked_dict = {}
    try:
        async_requests_get = _async_call(requests.get, timeout=5)
        data, thread = async_requests_get(pypi_url, timeout=3)
        if not data or isinstance(data, Exception):
            return None
        data = data.json()
        latest_version = data["info"]["version"]
        release_list = data["releases"].keys()
        for version, fields in data["releases"].items():
            for item in fields:
                yanked = item.get("yanked")
                yanked_reason = item.get("yanked_reason")
                if yanked:
                    yanked_dict[version] = yanked_reason
    except Exception:
        # Any issues whatsoever, just skip the latest version check.
        return None

    # Return if no update is available
    pip_prerelease = False
    deleted = False
    yanked = False
    yanked_reason = None
    parsed_current_version = _parse_version(current_version)

    # Check if current version has been yanked or deleted
    # NOTE: we will not return yanked or deleted if there is nothing to upgrade to
    if current_version in release_list:
        yanked = current_version in yanked_dict
        yanked_reason = yanked_dict.get(current_version)
    else:
        deleted = True

    # Check pre-releases
    if _parse_version(latest_version) <= parsed_current_version:
        # pre-releases are not included in latest_version
        # so if we are currently running a pre-release we check more
        if not parsed_current_version.is_prerelease:
            return None
        # Candidates are pre-releases with the same base_version
        release_list = map(_parse_version, release_list)
        release_list = filter(lambda v: v.is_prerelease, release_list)
        release_list = filter(
            lambda v: v.base_version == parsed_current_version.base_version,
            release_list,
        )
        release_list = sorted(release_list)
        if not release_list:
            return None

        parsed_latest_version = release_list[-1]
        if parsed_latest_version <= parsed_current_version:
            return None
        latest_version = str(parsed_latest_version)
        pip_prerelease = True

    return latest_version, pip_prerelease, deleted, yanked, yanked_reason



from __future__ import annotations

import pytest
from typing_extensions import (
    NotRequired,
    TypedDict,
)

from weave.legacy.weave import infer_types
from weave.legacy.weave import weave_types as types


class _TestNotRequiredTypedDict(TypedDict):
    a: NotRequired[int]
    b: int


def test_optional_typeddict_keys():
    assign_to = infer_types.python_type_to_type(_TestNotRequiredTypedDict)

    # Missing optional key ok
    assert assign_to.assign_type(types.TypedDict({"b": types.Int(), "c": types.Int()}))

    # Missing required key not ok
    assert not assign_to.assign_type(types.TypedDict())

    # Optional key wrong type not ok
    assert not assign_to.assign_type(
        types.TypedDict({"a": types.String(), "b": types.Int(), "c": types.Int()})
    )



import inspect
import typing

from weave.legacy.weave import infer_types, op_args

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types

InputTypeItemType = typing.Union[types.Type, typing.Callable[..., types.Type]]
InputTypeType = typing.Union[op_args.OpArgs, typing.Mapping[str, InputTypeItemType]]
MaybeInputTypeType = typing.Optional[InputTypeType]


def determine_input_type(
    pyfunc: typing.Callable,
    expected_input_type: MaybeInputTypeType = None,
    # TODO: I really don't like this boolean flag here.
    allow_unknowns: bool = False,
) -> op_args.OpArgs:
    python_input_type = _get_type_hints(pyfunc)
    if "return" in python_input_type:
        python_input_type.pop("return")

    declared_input_type = expected_input_type
    if declared_input_type is None:
        declared_input_type = {}
    weave_input_type = _create_args_from_op_input_type(declared_input_type)

    if isinstance(weave_input_type, op_args.OpNamedArgs):
        sig = get_signature(pyfunc)
        sig_params = sig.parameters
        arg_names = sig_params.keys()
        var_arg_params = [
            param
            for param in sig_params.values()
            if param.kind == param.VAR_POSITIONAL or param.kind == param.VAR_KEYWORD
        ]
        is_var_args = len(var_arg_params) > 0

        # validate there aren't extra declared Weave types
        if not is_var_args:
            weave_type_extra_arg_names = set(weave_input_type.arg_types.keys()) - set(
                arg_names
            )
            if weave_type_extra_arg_names:
                raise errors.WeaveDefinitionError(
                    "Weave Types declared for non-existent args: %s."
                    % weave_type_extra_arg_names
                )

        arg_types = {}

        # iterate through function args in order. The function determines
        # the order of arg_types, which is relied on everywhere.
        for input_name in arg_names:
            param = sig_params[input_name]
            if (
                param.kind == param.VAR_POSITIONAL
                or param.kind == param.VAR_KEYWORD
                or input_name == "_run"
            ):
                continue

            python_type = python_input_type.get(input_name)
            existing_weave_type = weave_input_type.arg_types.get(input_name)
            if python_type is not None:
                inferred_type = infer_types.python_type_to_type(python_type)
                if not allow_unknowns and inferred_type == types.UnknownType():
                    raise errors.WeaveDefinitionError(
                        "Weave Type could not be determined from Python type (%s) for arg: %s."
                        % (python_type, input_name)
                    )
                if existing_weave_type and existing_weave_type != inferred_type:
                    raise errors.WeaveDefinitionError(
                        "Python type (%s) and Weave Type (%s) declared for arg: %s. Remove one of them to fix."
                        % (inferred_type, existing_weave_type, input_name)
                    )
                arg_types[input_name] = inferred_type
            elif existing_weave_type:
                arg_types[input_name] = existing_weave_type
            else:
                arg_types[input_name] = types.UnknownType()

        if is_var_args:
            # Add any remaining weave types if we have var args
            for key, t in weave_input_type.arg_types.items():
                if key not in arg_names:
                    arg_types[key] = t

        # validate there aren't missing Weave types
        unknown_type_args = set(
            arg_name for arg_name, at in arg_types.items() if at == types.UnknownType()
        )
        weave_type_missing_arg_names = unknown_type_args - {"self"}
        if weave_type_missing_arg_names and not allow_unknowns:
            raise errors.WeaveDefinitionError(
                "Missing Weave Types for args: %s." % weave_type_missing_arg_names
            )

        weave_input_type = op_args.OpNamedArgs(arg_types)
    else:
        # TODO: no validation here...
        pass

    return weave_input_type


OutputTypeType = typing.Union[types.Type, typing.Callable[..., types.Type]]
MaybeOutputTypeType = typing.Optional[OutputTypeType]


def determine_output_type(
    pyfunc: typing.Callable,
    expected_output_type: MaybeOutputTypeType = None,
    allow_unknowns: bool = False,
) -> typing.Union[types.Type, typing.Callable[..., types.Type]]:
    type_hints = _get_type_hints(pyfunc)
    python_return_type = type_hints.get("return")
    inferred_output_type: types.Type
    if python_return_type is None:
        inferred_output_type = types.UnknownType()
    else:
        inferred_output_type = infer_types.python_type_to_type(python_return_type)
        if inferred_output_type == types.UnknownType():
            raise errors.WeaveDefinitionError(
                "Could not infer Weave Type from declared Python return type: %s"
                % python_return_type
            )

    weave_output_type = expected_output_type
    if weave_output_type is None:
        # weave output type is not declared, use type inferred from Python
        weave_output_type = inferred_output_type
    else:
        # Weave output_type was declared. Ensure compatibility with Python type.
        if callable(weave_output_type):
            if inferred_output_type != types.UnknownType():
                raise errors.WeaveDefinitionError(
                    "output_type is function but Python return type also declared. This is not yet supported"
                )
        elif (
            inferred_output_type != types.UnknownType()
            and not weave_output_type.assign_type(inferred_output_type)
        ):
            raise errors.WeaveDefinitionError(
                "Python return type not assignable to declared Weave output_type: %s !-> %s"
                % (inferred_output_type, weave_output_type)
            )
    if (
        not callable(weave_output_type)
        and weave_output_type == types.UnknownType()
        and not allow_unknowns
    ):
        raise errors.WeaveDefinitionError(
            "Op's return type must be declared: %s" % pyfunc
        )

    return weave_output_type


def get_signature(pyfunc: typing.Callable) -> inspect.Signature:
    if hasattr(pyfunc, "sig"):
        return pyfunc.sig  # type: ignore
    return inspect.signature(pyfunc)


def _create_args_from_op_input_type(input_type: InputTypeType) -> op_args.OpArgs:
    if isinstance(input_type, op_args.OpArgs):
        return input_type
    if not isinstance(input_type, dict):
        raise errors.WeaveDefinitionError("input_type must be OpArgs or a dict")
    for k, v in input_type.items():
        if not isinstance(v, types.Type) and not callable(v):
            raise errors.WeaveDefinitionError(
                "input_type must be dict[str, Type] but %s is %s" % (k, v)
            )
    return op_args.OpNamedArgs(input_type)


def _get_type_hints(pyfunc: typing.Callable) -> dict[str, typing.Any]:
    if hasattr(pyfunc, "sig"):
        type_hints = {
            k: p.annotation
            for k, p in pyfunc.sig.parameters.items()  # type: ignore
            if p.annotation is not p.empty
        }
    else:
        type_hints = typing.get_type_hints(pyfunc)
    return type_hints



import base64
import hashlib
import typing

from . import refs_internal

TRACE_REF_SCHEME = "weave"
ARTIFACT_REF_SCHEME = "wandb-artifact"
WILDCARD_ARTIFACT_VERSION_AND_PATH = ":*"


def bytes_digest(json_val: bytes) -> str:
    hasher = hashlib.sha256()
    hasher.update(json_val)
    hash_bytes = hasher.digest()
    base64_encoded_hash = base64.urlsafe_b64encode(hash_bytes).decode("utf-8")
    return base64_encoded_hash.replace("-", "X").replace("_", "Y").rstrip("=")


def str_digest(json_val: str) -> str:
    return bytes_digest(json_val.encode())


def _order_dict(dictionary: typing.Dict) -> typing.Dict:
    return {
        k: _order_dict(v) if isinstance(v, dict) else v
        for k, v in sorted(dictionary.items())
    }


def encode_bytes_as_b64(contents: typing.Dict[str, bytes]) -> typing.Dict[str, str]:
    res = {}
    for k, v in contents.items():
        if isinstance(v, bytes):
            res[k] = base64.b64encode(v).decode("ascii")
        else:
            raise ValueError(f"Unexpected type for file {k}: {type(v)}")
    return res


def decode_b64_to_bytes(contents: typing.Dict[str, str]) -> typing.Dict[str, bytes]:
    res = {}
    for k, v in contents.items():
        if isinstance(v, str):
            res[k] = base64.b64decode(v.encode("ascii"))
        else:
            raise ValueError(f"Unexpected type for file {k}: {type(v)}")
    return res


valid_schemes = [
    TRACE_REF_SCHEME,
    ARTIFACT_REF_SCHEME,
    refs_internal.WEAVE_INTERNAL_SCHEME,
]


def extract_refs_from_values(
    vals: typing.Any,
) -> typing.List[str]:
    refs = []

    def _visit(val: typing.Any) -> typing.Any:
        if isinstance(val, dict):
            for v in val.values():
                _visit(v)
        elif isinstance(val, list):
            for v in val:
                _visit(v)
        elif isinstance(val, str) and any(
            val.startswith(scheme + "://") for scheme in valid_schemes
        ):
            refs.append(val)

    _visit(vals)
    return refs


def assert_non_null_wb_user_id(obj: typing.Any) -> None:
    if not hasattr(obj, "wb_user_id") or obj.wb_user_id is None:
        raise ValueError("wb_user_id cannot be None")


def assert_null_wb_user_id(obj: typing.Any) -> None:
    if hasattr(obj, "wb_user_id") and obj.wb_user_id is not None:
        raise ValueError("wb_user_id must be None")



from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType


def op_get_tag_type_resolver(obj_type: types.Type) -> types.Type:
    if isinstance(obj_type, TaggedValueType) or (
        isinstance(obj_type, types.Const)
        and isinstance(obj_type.val_type, TaggedValueType)
    ):
        return obj_type.tag
    else:
        return types.NoneType()


def op_make_type_tagged_resolver(
    obj_type: types.Type, tag_type: types.Type
) -> types.Type:
    if isinstance(tag_type, types.TypedDict):
        return TaggedValueType(tag_type, obj_type)
    else:
        return obj_type


def op_make_type_key_tag_resolver(
    obj_type: types.Type, key: str, tag_type: types.Type
) -> types.Type:
    tags = {key: tag_type}
    if isinstance(tag_type, TaggedValueType) and isinstance(
        tag_type.tag, types.TypedDict
    ):
        tags.update(tag_type.tag.property_types)
    return TaggedValueType(types.TypedDict(tags), obj_type)



import random
import typing

import pytest

import weave
from weave.legacy.weave import ops_arrow

from weave.legacy.tests.util.concrete_tagged_value import (
    TaggedValue,
    concrete_from_tagstore,
    concrete_to_tagstore,
)


class ObjSubFields(typing.TypedDict):
    x: str
    y: str
    z: int


@weave.type()
class _TestMyObj:
    a: ObjSubFields
    b: float


def assert_equal(v1, v2):
    if isinstance(v1, dict):
        # Allow missing key to be None
        assert isinstance(v2, dict)
        all_keys = set(v1.keys()) | set(v2.keys())
        for k in all_keys:
            if k not in v1:
                assert v2[k] == None
            elif k not in v2:
                assert v1[k] == None
            else:
                assert_equal(v1[k], v2[k])
    elif isinstance(v1, list):
        assert isinstance(v2, list)
        assert len(v1) == len(v2)
        for i in range(len(v1)):
            assert_equal(v1[i], v2[i])
    elif isinstance(v1, ops_arrow.ArrowWeaveList):
        assert isinstance(v2, ops_arrow.ArrowWeaveList)
        assert v1.object_type == v2.object_type
        assert v1._arrow_data == v2._arrow_data
    elif isinstance(v1, TaggedValue):
        assert isinstance(v2, TaggedValue)
        assert_equal(v1.value, v2.value)
        assert_equal(v1.tag, v2.tag)
    else:
        assert v1 == v2


@pytest.mark.parametrize(
    "value",
    [
        [15, 2],
        ["a", "b"],
        ["a", 15],
        ["a", None],
        ["a", 15, None],
        [{"a": 15, "b": 2}, {"a": 3, "b": 4}],
        [{"a": 15, "b": [8, 9, 10]}, {"a": 3, "b": [4, 5, 6]}],
        [
            {"a": 15, "b": ops_arrow.to_arrow([8, 9, 10])},
            {"a": 3, "b": ops_arrow.to_arrow([4, 5, 6])},
        ],
        [{"a": 15, "b": [{"c": 5}, {"c": -2}]}, {"a": 3, "b": [{"c": -2}, {"c": 4}]}],
        # basic tag array
        [
            TaggedValue({"x": 99}, 3),
            TaggedValue({"x": 101}, 4),
        ],
        # tagged tag
        [
            TaggedValue({"a": 1, "b": TaggedValue({"x": 99}, 3)}, 1),
            TaggedValue({"a": 2, "b": TaggedValue({"x": 101}, 4)}, 2),
        ],
        # Tagged union
        [
            TaggedValue({"x": 99}, "a"),
            TaggedValue({"x": 101}, 4),
        ],
        # tagged value array with tagged union for tags
        [
            TaggedValue({"a": 1, "b": TaggedValue({"x": 99}, "a")}, 1),
            TaggedValue({"a": 2, "b": TaggedValue({"x": 101}, 4)}, 2),
        ],
        [
            TaggedValue({"x": 99}, [None, TaggedValue({"j": 5}, 99)]),
        ],
        [
            TaggedValue(
                {"a": 1, "b": TaggedValue({"x": 99}, 3)},
                {
                    "a": [None, TaggedValue({"j": 5}, -14)],
                    "b": [None, TaggedValue({"yy": -10}, -100)],
                },
            )
        ],
        [
            TaggedValue(
                {"a": 1, "b": TaggedValue({"x": 99}, 3)},
                {
                    "a": [
                        None,
                        TaggedValue(
                            {"j": _TestMyObj({"x": "xxx", "y": "yyy", "z": 5}, 14.1)},
                            -14,
                        ),
                    ],
                    "b": [
                        None,
                        TaggedValue(
                            {
                                "yy": _TestMyObj(
                                    {"x": "xx5", "y": "yy5", "z": 1024}, 0.0001
                                )
                            },
                            -100,
                        ),
                    ],
                },
            )
        ],
        ["a", 15, TaggedValue({"a": [1, "a", None]}, [None, "b", 6])],
        [
            "a",
            15,
            TaggedValue(
                {"a": [1, "a", None, TaggedValue({"j": "x"}, [55, "xx", None])]},
                [None, "b", 6, TaggedValue({"k": "k"}, [None, "c", 7])],
            ),
        ],
        [
            "a",
            15,
            TaggedValue(
                {
                    "a": [
                        1,
                        "a",
                        None,
                        TaggedValue(
                            {
                                "j": _TestMyObj(
                                    {"x": "aaa", "y": "aba", "z": -11111}, 1.23
                                )
                            },
                            [55, "xx", None],
                        ),
                    ]
                },
                [
                    None,
                    "b",
                    6,
                    TaggedValue(
                        {"k": "k"},
                        [
                            None,
                            {"x": 5},
                            {"a": 1.249124},
                            _TestMyObj({"x": "xx", "y": "yy", "z": -1}, 515.12),
                            7,
                        ],
                    ),
                ],
            ),
        ],
    ],
)
def test_to_from_arrow(value):
    weave_py = concrete_to_tagstore(value)
    a = ops_arrow.to_arrow(weave_py)

    leaf_col_paths = []

    def save_leaf(awl, path):
        if isinstance(awl.object_type, weave.types.String):
            leaf_col_paths.append(path)

    a.map_column(save_leaf)
    print("LEAF COL PATHS", leaf_col_paths)

    test_ids = range(2 ** len(leaf_col_paths))
    if len(test_ids) > 10:
        test_ids = random.sample(test_ids, 10)
    for test_id in test_ids:
        f_string = "{:0" + str(len(leaf_col_paths)) + "b}"
        bits = f_string.format(test_id)
        dict_cols = []
        for bit, path in zip(bits, leaf_col_paths):
            if bit == "1":
                dict_cols.append(path)
                a = a.replace_column(path, lambda v: v.dictionary_encode())
        print("TEST ", test_id, "dict_col_mask:", bits, "dict_col_paths:", dict_cols)
        print("A", a)
        print("A._arrow_data", a._arrow_data.type)
        weave_py2 = a.to_pylist_tagged()
        value2 = concrete_from_tagstore(weave_py2)
        print("  VALUE 1", value)
        print("  VALUE 2", value2)
        assert_equal(value, value2)
        # assert value == value2


# TODO: test tags, unions



import weave
from weave.legacy.weave import graph, weave_internal
from weave.legacy.weave import weave_types as types


def test_map_dag_produces_same_len():
    #         3
    #          \
    #   (a) -> (b)
    #     \      \
    # 4 -> (c) -> (d)

    a = weave_internal.make_var_node(types.Int(), "a")
    b = a + 3
    c = a + 4
    d = b + c

    assert id(d.from_op.inputs["lhs"].from_op.inputs["lhs"]) == id(
        d.from_op.inputs["rhs"].from_op.inputs["lhs"]
    )

    assert graph.count(d) == 6

    def replace_a(node: graph.Node) -> graph.Node:
        if isinstance(node, graph.VarNode):
            return weave_internal.make_var_node(types.Int(), "b")
        return node

    mapped_d = graph.map_nodes_top_level([d], replace_a)[0]
    assert id(mapped_d.from_op.inputs["lhs"].from_op.inputs["lhs"]) == id(
        mapped_d.from_op.inputs["rhs"].from_op.inputs["lhs"]
    )
    assert graph.count(mapped_d) == 6


def test_map_nodes_toplevel_doesnt_walk_lambdas():
    l = weave.save([1, 2, 3])
    node = l.map(lambda x: x + 1)
    node_count = {"count": 0}

    def _map_fn(node):
        node_count["count"] += 1

    graph.map_nodes_top_level([node], _map_fn)[0]
    assert node_count["count"] == 4


def test_map_nodes_full_walks_lambdas():
    l = weave.save([1, 2, 3])
    node = l.map(lambda x: x + 1)
    node_count = {"count": 0}

    def _map_fn(node):
        node_count["count"] += 1

    graph.map_nodes_full([node], _map_fn)[0]
    assert node_count["count"] == 7


def test_map_nodes_full_replaces_in_lambda():
    l = weave.save([1, 2, 3])
    node = l.map(lambda x: x + 1)
    assert weave.use(node) == [2, 3, 4]

    def _map_fn(node):
        if isinstance(node, graph.OutputNode) and node.from_op.name == "number-add":
            return graph.OutputNode(node.type, "number-mult", node.from_op.inputs)

    result = graph.map_nodes_full([node], _map_fn)[0]
    assert weave.use(result) == [1, 2, 3]


def test_linearize():
    a = weave_internal.make_var_node(types.Int(), "a")
    dag = ((a + 1) * 2) + 3
    linear = graph.linearize(dag)
    assert len(linear) == 3
    assert linear[0].from_op.name == "number-add"
    assert list(linear[0].from_op.inputs.values())[1].val == 1
    assert linear[1].from_op.name == "number-mult"
    assert list(linear[1].from_op.inputs.values())[1].val == 2
    assert linear[2].from_op.name == "number-add"
    assert list(linear[2].from_op.inputs.values())[1].val == 3


def test_replace_node():
    #    2      3
    #     \      \
    # 1 -> (a) -> (b)
    #        \      \
    #   4 -> (c) -> (d)
    a = weave_internal.make_const_node(types.Int(), 1) + 2
    b = a + 3
    c = a * 4
    d = b + c
    assert weave.use(d) == 18

    def replace_c(node):
        if node is c:
            return c.from_op.inputs["lhs"] / 4

    x = graph.map_nodes_top_level([d], replace_c)[0]
    assert weave.use(x) == 6.75



import dataclasses
import json
import re
import typing

import numpy as np
import pytest
import wandb

from weave.legacy.weave import api as weave
from weave.legacy.weave import (
    artifact_mem,
    artifact_wandb,
    mappers_python,
    ops_arrow,
    storage,
)
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.arrow import list_ as arrow
from weave.legacy.weave.weave_internal import make_const_node

from ...legacy.weave.weavejs_fixes import recursively_unwrap_unions
from . import test_helpers


@weave.type()
class SomeObj:
    some_int: int
    some_str: str


@dataclasses.dataclass
class SomeCustomObj:
    obj: typing.Any


class SomeCustomObjType(types.Type):
    name = "test_storage_somecustomobj"
    instance_classes = SomeCustomObj
    instance_class = SomeCustomObj

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.json") as f:
            json.dump(obj.obj, f)

    @classmethod
    def load_instance(cls, artifact, name, extra=None):
        with artifact.open(f"{name}.json") as f:
            return SomeCustomObj(json.load(f))


def test_dict():
    obj = {"x": 14}
    obj_id = storage.save(obj, "my-dict")
    obj2 = storage.get(obj_id)
    assert obj == obj2


def test_nested_dict():
    obj = {"a": 1, "b": {"c": 2}}
    obj_id = storage.save(obj, "my-dict")
    obj2 = storage.get(obj_id)
    assert obj == obj2


def test_doubly_nested_dict():
    np_array = np.array([[4, 5], [6, 7]])
    obj = {"a": np_array, "b": {"c": np_array, "d": 3}}
    obj_id = storage.save(obj, "my-dict")
    obj2 = storage.get(obj_id)
    assert obj.keys() == obj2.keys()
    assert np.array_equal(obj["a"], obj2["a"])
    assert np.array_equal(obj["b"]["c"], obj2["b"]["c"])
    assert obj["b"]["d"] == obj2["b"]["d"]


def test_list_with_arrays():
    obj = [{"a": np.array([4, 5]), "b": "b0"}, {"a": np.array([9, 10]), "b": "b1"}]
    obj_id = storage.save(obj, "my-list-with-arrays")
    obj2 = storage.get(obj_id)
    assert np.array_equal(obj[0]["a"], obj2[0]["a"])
    assert np.array_equal(obj[1]["a"], obj2[1]["a"])
    assert obj[0]["b"] == obj[0]["b"]
    assert obj[1]["b"] == obj[1]["b"]


def test_numpy():
    np_array = np.array([[4, 5], [6, 7]])
    obj_id = storage.save(np_array, "my-arr")
    np_array2 = storage.get(obj_id)
    assert np.array_equal(np_array, np_array2)


def test_custom_obj():
    obj = SomeCustomObj({"a": 5})
    obj_id = storage.save(obj, "my-obj")
    obj2 = storage.get(obj_id)
    assert obj.obj == obj2.obj


@pytest.mark.skip(reason="wb table doesnt work right now")
def test_wandb_table():
    obj = SomeCustomObj({"a": 5})
    obj_id = storage.save(obj, "my-obj")
    obj2 = storage.get(obj_id)
    assert obj.obj == obj2.obj
    table = wandb.Table(["x", "y"])
    table.add_data(1, 2)
    table_id = storage.save(table, "my-table")
    table2 = storage.get(table_id)
    assert table._eq_debug(table2)


def test_cross_obj_ref():
    d1 = {"a": 5, "b": SomeCustomObj(14)}
    d1_id = storage.save(d1, "my-d1")
    d2 = storage.get(d1_id)
    d3 = {"f": 5, "c": d2["b"]}
    d3_id = storage.save(d3, "my-d3")
    d4 = storage.get(d3_id)
    assert d4.keys() == d3.keys()
    assert d3["f"] == d4["f"]
    assert d3["c"].obj == d4["c"].obj


def test_cross_obj_outer_ref():
    d1 = {"a": 5, "b": SomeCustomObj(14)}
    d1_id = storage.save(d1, "my-d1")
    d2 = storage.get(d1_id)
    d3 = {"f": 5, "c": d2}
    d3_id = storage.save(d3, "my-d3")
    d4 = storage.get(str(d3_id))
    assert d4.keys() == d3.keys()
    assert d3["f"] == d4["f"]
    assert d3["c"]["a"] == d4["c"]["a"]
    assert d3["c"]["b"].obj == d4["c"]["b"].obj


def test_ref_to_item_in_list():
    l = [{"a": 5, "b": 6}]
    l_node = weave.save(l, "my-l")
    l_node = l_node[0]["a"]

    dict_with_ref = {"c": l_node}
    d_node = weave.save(dict_with_ref, "my-dict_with_ref")

    assert weave.use(d_node["c"] == 5) == True


def test_list_of_ref_to_item_in_list():
    l = [{"a": 5, "b": 6}, {"a": 7, "b": 9}]
    l_node = weave.save(l, "my-l")

    list_dict_with_ref = [{"c": l_node[0]["a"]}, {"c": l_node[1]["a"]}]
    d_node = weave.save(list_dict_with_ref, "my-dict_with_ref")

    assert weave.use(d_node[0]["c"] == 5) == True
    assert weave.use(d_node[1]["c"] == 7) == True


def test_ref_type(test_artifact_dir):
    obj = {"x": 14}
    ref = storage.save(obj, "my-dict")
    python_ref = storage.to_python(ref)
    assert python_ref == {
        "_type": {
            "type": "LocalArtifactRef",
            "_base_type": {
                "type": "FilesystemArtifactRef",
                "_base_type": {"type": "Ref"},
            },
            "objectType": {"type": "typedDict", "propertyTypes": {"x": "int"}},
        },
        "_val": "local-artifact:///my-dict:10e1804d2dd19195ac2d/obj",
    }
    ref2 = storage.from_python(python_ref)
    obj2 = storage.deref(ref2)
    assert obj == obj2


def test_boxing():
    ref = storage.save(5)
    val = storage.get(str(ref))
    assert val == 5
    assert val._ref is not None


def test_saveload_type():
    t = types.TypedDict({"a": types.Int(), "b": types.String()})
    t_type = types.TypeRegistry.type_of(t)
    ref = storage.save(t)
    t2 = ref.get()
    assert t == t2


def test_list_obj():
    list_obj = [SomeObj(1, "a"), SomeObj(2, "b")]
    ref = storage.save(list_obj, "my-listobj")
    list_obj2 = storage.get(ref)
    assert list_obj[0].some_int == list_obj2[0].some_int
    assert list_obj[0].some_str == list_obj2[0].some_str
    assert list_obj[1].some_int == list_obj2[1].some_int
    assert list_obj[1].some_str == list_obj2[1].some_str


def test_cross_artifact_ref():
    owner = {"a": 1, "b": {"c": SomeCustomObj(2)}}
    node = weave.save(owner, "owner-obj")
    assert weave.use(node["b"]) == {"c": SomeCustomObj(2)}
    # TODO: assert that ref is to original object


@weave.type()
class _TestObjType:
    pass


def test_ref_to_node():
    d = weave.save({"a": _TestObjType()})
    node = d["a"]
    # Node a Node with weave type ObjectType
    # We want to make sure dispatch doesn't convert this access to ._ref into
    # a getattr op call (prevented by ref=None on FallbackNodeTypeDispatcherMixin)
    # ref should be None here
    assert node._ref == None


@pytest.mark.parametrize(
    "obj, wb_type, expected",
    [
        ([], None, {"_type": {"type": "list", "objectType": "unknown"}, "_val": []}),
        (
            [{"a": []}, {"a": [1]}],
            None,
            {
                "_type": {
                    "type": "list",
                    "objectType": {
                        "type": "typedDict",
                        "propertyTypes": {"a": {"type": "list", "objectType": "int"}},
                    },
                },
                "_val": [{"a": []}, {"a": [1]}],
            },
        ),
        (
            [{"a": []}, {"a": [1]}, {"a": [None]}],
            None,
            {
                "_type": {
                    "type": "list",
                    "objectType": {
                        "type": "typedDict",
                        "propertyTypes": {
                            "a": {
                                "type": "list",
                                "objectType": {
                                    "type": "union",
                                    "members": ["int", "none"],
                                },
                            }
                        },
                    },
                },
                "_val": [{"a": []}, {"a": [1]}, {"a": [None]}],
            },
        ),
    ],
)
def test_to_python_object(obj, wb_type, expected):
    assert recursively_unwrap_unions(storage.to_python(obj, wb_type)) == expected


class MockStats:
    call_count = 0


@pytest.fixture
def mock_get_wandb_read_artifact_uri():
    orig = artifact_wandb.get_wandb_read_artifact_uri

    stats = MockStats()

    def _mock_get_wandb_read_artifact_uri(path: str):
        stats.call_count += 1
        return artifact_wandb.WeaveWBArtifactURI(
            "my-uri", "a_specific_version", "entity", "project", None, "path"
        )

    artifact_wandb.get_wandb_read_artifact_uri = _mock_get_wandb_read_artifact_uri
    yield stats
    artifact_wandb.get_wandb_read_artifact_uri = orig


def test_to_conversion_uri_resolution(
    mock_get_wandb_read_artifact_uri: MockStats,
):
    def make_art():
        return artifact_wandb.WandbArtifact(
            "my-art",
            "my-type",
            artifact_wandb.WeaveWBArtifactURI(
                "my-uri", "latest", "entity", "project", None, "path"
            ),
        )

    wb_type = types.TypeRegistry.type_of(make_art())

    # Construct the mapper manually. Default should be to resolve
    mapper = mappers_python.map_to_python(wb_type, artifact_mem.MemArtifact())
    res = mapper.apply(make_art())
    assert mock_get_wandb_read_artifact_uri.call_count == 1
    assert res == "wandb-artifact:///entity/project/my-uri:a_specific_version"

    # to_python uses the resolve behavior
    res = storage.to_python(make_art())
    assert mock_get_wandb_read_artifact_uri.call_count == 2
    assert res == {
        "_type": {
            "_base_type": {"type": "FilesystemArtifact"},
            "type": "WandbArtifact",
        },
        "_val": "wandb-artifact:///entity/project/my-uri:a_specific_version",
    }

    # Tell the mapper not to resolve
    mapper = mappers_python.map_to_python(
        wb_type, artifact_mem.MemArtifact(), mapper_options={"use_stable_refs": False}
    )
    assert mock_get_wandb_read_artifact_uri.call_count == 2
    res = mapper.apply(make_art())
    assert res == "wandb-artifact:///entity/project/my-uri:latest"

    # to_weavejs doesn't resolve
    assert mock_get_wandb_read_artifact_uri.call_count == 2
    js_serializer = storage.make_js_serializer()
    res = js_serializer(make_art())
    assert res == "wandb-artifact:///entity/project/my-uri:latest"


def test_save_to_branch():
    data = [{"a": 5}, {"a": 6}]
    ref = storage.save(data, "data:main")
    assert ref.branch == "main"
    data2 = storage.get("local-artifact:///data:main/obj")
    assert data == data2


def test_branch_point():
    data = [{"a": 5}, {"a": 6}]
    ref = storage.save(data, "data:main")
    assert ref.branch == "main"
    data2 = storage.get("local-artifact:///data:main/obj")
    data2[0]["a"] = 7
    ref2 = storage.save(data2, "data:main", branch="other")
    assert ref2.branch == "other"
    assert ref2.branch_point["original_uri"] == "local-artifact:///data:main"
    assert ref2.branch_point["commit"] == ref.version

    data3 = storage.get("local-artifact:///data:other/obj")
    data3[0]["a"] = 8
    ref3 = storage.save(data3, "data:other")
    assert ref3.branch == "other"
    # Still points to main branch point since we didn't fork
    assert ref3.branch_point["original_uri"] == "local-artifact:///data:main"
    assert ref3.branch_point["commit"] == ref.version

    data4 = storage.get("local-artifact:///data:other/obj")
    data4[0]["a"] = 8
    ref4 = storage.save(data4, "data:other", branch="other2")
    assert ref4.branch == "other2"
    assert ref4.branch_point["original_uri"] == "local-artifact:///data:other"
    assert ref4.branch_point["commit"] == ref3.version


def test_to_python_norefs():
    data = [{"a": 5}, {"a": 6}]
    result = storage.to_python(data)
    assert result["_val"] == data

    assert len(storage.local_artifacts()) == 0
    data2 = storage.from_python(result)
    assert data == data2


def test_to_python_ref():
    custom_obj = ops_arrow.to_arrow([{"x": 14}])
    data = [{"a": 5}, {"a": custom_obj}]
    result = storage.to_python(data)

    # There should be one artifact
    assert len(storage.local_artifacts()) == 1

    # The result should have a ref into that artifact
    assert result["_val"] == [
        {"a": {"_union_id": 0, "_val": 5}},
        {
            "a": {
                "_union_id": 1,
                "_val": test_helpers.RegexMatcher("local-artifact:///list:.*/a/0"),
            }
        },
    ]

    data2 = storage.from_python(result)
    assert data[0] == data2[0]
    assert data[1]["a"].to_pylist_raw() == data2[1]["a"].to_pylist_raw()


def test_to_python_toplevel_ref():
    data = ops_arrow.to_arrow([{"x": 14}])
    result = storage.to_python(data)
    assert result["_val"] == test_helpers.RegexMatcher(
        "local-artifact:///ArrowWeaveList:.*/0"
    )

    data2 = storage.from_python(result)
    assert data.to_pylist_raw() == data2.to_pylist_raw()



import pytest

from weave.legacy.weave import graph, ops, weave_internal
from weave.legacy.weave import weave_types as types

from ...legacy.weave import weavify
from weave.legacy.tests.util import geom


@pytest.mark.parametrize(
    "name,object,expected",
    [
        ("dict", {"a": 1, "b": 2}, ops.dict_(**{"a": 1, "b": 2})),
        ("int", 1, weave_internal.make_const_node(types.Int(), 1)),
        ("float", 1.0, weave_internal.make_const_node(types.Float(), 1.0)),
        ("string", "a", weave_internal.make_const_node(types.String(), "a")),
        ("bool", True, weave_internal.make_const_node(types.Boolean(), True)),
        (
            "mixed list",
            [1, weave_internal.make_const_node(types.Int(), 2)],
            ops.make_list(**{"0": 1, "1": 2}),
        ),
        (
            "recursive dict",
            {"a": {"b": 1}, "d": 2},
            ops.dict_(**{"a": ops.dict_(**{"b": 1}), "d": 2}),
        ),
        (
            "recursive list",
            [1, [2, 3]],
            ops.make_list(**{"0": 1, "1": ops.make_list(**{"0": 2, "1": 3})}),
        ),
        (
            "recursive mixed",
            {"a": [1, 2], "b": {"c": 3}},
            ops.dict_(
                **{"a": ops.make_list(**{"0": 1, "1": 2}), "b": ops.dict_(**{"c": 3})}
            ),
        ),
        ("none", None, weave_internal.make_const_node(types.NoneType(), None)),
        # decorator_type() constructor generation is disabled.
        # (
        #     "object",
        #     geom.Point2d(1, 2),
        #     geom.Point2d.constructor(x=1, y=2),  # type: ignore
        # ),
        # (
        #     "mixed object",
        #     {"a": geom.Point2d(1, 2)},
        #     ops.dict_(**{"a": geom.Point2d.constructor(x=1, y=2)}),  # type: ignore
        # ),
        # (
        #     "mixed object list",
        #     [
        #         geom.Point2d(1, 2),
        #         geom.Point2d.constructor(  # type: ignore
        #             x=weave_internal.make_const_node(types.Int(), 1),
        #             y=weave_internal.make_const_node(types.Int(), 2),
        #         ),
        #     ],
        #     ops.make_list(
        #         **{
        #             "0": geom.Point2d.constructor(x=1, y=2),  # type: ignore
        #             "1": geom.Point2d.constructor(x=1, y=2),  # type: ignore
        #         }
        #     ),
        # ),
    ],
)
def test_weavify_object(name, object, expected):
    assert graph.nodes_equal(weavify.weavify_object(object), expected)
